# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

from tvb.simulator.integrators import *
from tvb.adapters.simulator.noise_forms import get_ui_name_to_noise_dict
from tvb.core.neotraits.forms import Form, ScalarField, SimpleSelectField


def get_integrator_to_form_dict():
    integrator_class_to_form = {
        HeunDeterministic: HeunDeterministicIntegratorForm,
        HeunStochastic: HeunStochasticIntegratorForm,
        EulerDeterministic: EulerDeterministicIntegratorForm,
        EulerStochastic: EulerStochasticIntegratorForm,
        RungeKutta4thOrderDeterministic: RungeKutta4thOrderDeterministicIntegratorForm,
        Identity: IdentityIntegratorForm,
        VODE: VODEIntegratorForm,
        VODEStochastic: VODEStochasticIntegratorForm,
        Dopri5: Dopri5IntegratorForm,
        Dopri5Stochastic: Dopri5StochasticIntegratorForm,
        Dop853: Dop853IntegratorForm,
        Dop853Stochastic: Dop853StochasticIntegratorForm
    }
    return integrator_class_to_form


def get_ui_name_to_integrator_dict():
    ui_name_to_integrator = {
        'Heun': HeunDeterministic,
        'Stochastic Heun': HeunStochastic,
        'Euler': EulerDeterministic,
        'Euler-Maruyama': EulerStochastic,
        'Runge-Kutta 4th order': RungeKutta4thOrderDeterministic,
        '"Difference equation': Identity,
        'Variable-order Adams / BDF': VODE,
        'Stochastic variable-order Adams / BDF': VODEStochastic,
        'Dormand-Prince, order (4, 5)': Dopri5,
        'Stochastic Dormand-Prince, order (4, 5)': Dopri5Stochastic,
        'Dormand-Prince, order 8 (5, 3)': Dop853,
        'Stochastic Dormand-Prince, order 8 (5, 3)': Dop853Stochastic,

    }
    return ui_name_to_integrator


def get_form_for_integrator(integrator_class):
    return get_integrator_to_form_dict().get(integrator_class)


class IntegratorForm(Form):

    def __init__(self, prefix=''):
        super(IntegratorForm, self).__init__(prefix)
        self.dt = ScalarField(Integrator.dt, self)


class IntegratorStochasticForm(IntegratorForm):
    template = 'form_fields/select_field.html'

    def __init__(self, prefix=''):
        super(IntegratorStochasticForm, self).__init__(prefix)
        # TODO: show select box with Noise types
        # self.noise = FormField(MultiplicativeNoiseForm, self, name='noise', label='Noise')
        self.noise_choices = get_ui_name_to_noise_dict()
        self.noise = SimpleSelectField(self.noise_choices, self, name='noise', required=True, label='Noise')

    def fill_trait(self, datatype):
        super(IntegratorStochasticForm, self).fill_trait(datatype)
        datatype.noise = self.noise.data()

    def fill_from_trait(self, trait):
        # type: (Integrator) -> None
        self.noise.data = trait.noise.__class__


class HeunDeterministicIntegratorForm(IntegratorForm):

    def __init__(self, prefix=''):
        super(HeunDeterministicIntegratorForm, self).__init__(prefix)


class HeunStochasticIntegratorForm(IntegratorStochasticForm):

    def __init__(self, prefix=''):
        super(HeunStochasticIntegratorForm, self).__init__(prefix)


class EulerDeterministicIntegratorForm(IntegratorForm):

    def __init__(self, prefix=''):
        super(EulerDeterministicIntegratorForm, self).__init__(prefix)


class EulerStochasticIntegratorForm(IntegratorStochasticForm):

    def __init__(self, prefix=''):
        super(EulerStochasticIntegratorForm, self).__init__(prefix)


class RungeKutta4thOrderDeterministicIntegratorForm(IntegratorForm):

    def __init__(self, prefix=''):
        super(RungeKutta4thOrderDeterministicIntegratorForm, self).__init__(prefix)


class IdentityIntegratorForm(IntegratorForm):

    def __init__(self, prefix=''):
        super(IdentityIntegratorForm, self).__init__(prefix)


class VODEIntegratorForm(IntegratorForm):

    def __init__(self, prefix=''):
        super(VODEIntegratorForm, self).__init__(prefix)


class VODEStochasticIntegratorForm(IntegratorStochasticForm):

    def __init__(self, prefix=''):
        super(VODEStochasticIntegratorForm, self).__init__(prefix)


class Dopri5IntegratorForm(IntegratorForm):

    def __init__(self, prefix=''):
        super(Dopri5IntegratorForm, self).__init__(prefix)


class Dopri5StochasticIntegratorForm(IntegratorStochasticForm):

    def __init__(self, prefix=''):
        super(Dopri5StochasticIntegratorForm, self).__init__(prefix)


class Dop853IntegratorForm(IntegratorForm):

    def __init__(self, prefix=''):
        super(Dop853IntegratorForm, self).__init__(prefix)


class Dop853StochasticIntegratorForm(IntegratorStochasticForm):

    def __init__(self, prefix=''):
        super(Dop853StochasticIntegratorForm, self).__init__(prefix)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Module in charge with Launching an operation (creating the Operation entity as well, based on gathered parameters).

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
.. moduleauthor:: Ionel Ortelecan <ionel.ortelecan@codemart.ro>
.. moduleauthor:: Yann Gordon <yann@tvb.invalid>
"""

import os
import json
import zipfile
import sys
from copy import copy
from cgi import FieldStorage
from tvb.adapters.simulator.simulator_adapter import SimulatorAdapter
from tvb.adapters.datatypes.db.time_series import TimeSeriesIndex
from tvb.adapters.analyzers.metrics_group_timeseries import TimeseriesMetricsAdapter, TimeseriesMetricsAdapterForm
from tvb.analyzers.metrics_base import BaseTimeseriesMetricAlgorithm
from tvb.basic.exceptions import TVBException
from tvb.basic.neotraits.api import Range
from tvb.basic.profile import TvbProfile
from tvb.basic.logger.builder import get_logger
from tvb.core import utils
from tvb.core.adapters import constants
from tvb.core.adapters.abcadapter import ABCAdapter, ABCSynchronous
from tvb.core.adapters.exceptions import LaunchException
from tvb.core.entities.model.model_burst import PARAM_RANGE_PREFIX, RANGE_PARAMETER_1, RANGE_PARAMETER_2, \
    BurstConfiguration
from tvb.core.entities.model.model_datatype import DataTypeGroup
from tvb.core.entities.model.model_operation import STATUS_FINISHED, STATUS_ERROR, OperationGroup, Operation
from tvb.core.entities.storage import dao
from tvb.core.entities.transient.structure_entities import DataTypeMetaData
from tvb.core.entities.file.files_helper import FilesHelper
from tvb.core.neocom import h5
from tvb.core.neotraits.h5 import ViewModelH5
from tvb.core.services.burst_service import BurstService
from tvb.core.services.backend_client import BACKEND_CLIENT
from tvb.core.services.simulator_serializer import SimulatorSerializer

try:
    from cherrypy._cpreqbody import Part
    # cover cases when the web interface is not available.
except Exception:
    Part = FieldStorage

TEMPORARY_PREFIX = ".tmp"

RANGE_PARAMETER_1 = RANGE_PARAMETER_1
RANGE_PARAMETER_2 = RANGE_PARAMETER_2

UIKEY_SUBJECT = "RESERVEDsubject"
UIKEY_USERGROUP = "RESERVEDusergroup"


class OperationService:
    """
    Class responsible for preparing an operation launch. 
    It will prepare parameters, and decide if the operation is to be executed
    immediately, or to be sent on the cluster.
    """
    ATT_UID = "uid"

    def __init__(self):
        self.logger = get_logger(self.__class__.__module__)
        self.file_helper = FilesHelper()

    ##########################################################################################
    ######## Methods related to launching operations start here ##############################
    ##########################################################################################

    def initiate_operation(self, current_user, project, adapter_instance, visible=True, model_view=None, **kwargs):
        """
        Gets the parameters of the computation from the previous inputs form,
        and launches a computation (on the cluster or locally).
        
        Invoke custom method on an Adapter Instance. Make sure when the  
        operation has finished that the correct results are stored into DB. 
        """
        if not isinstance(adapter_instance, ABCAdapter):
            self.logger.warning("Inconsistent Adapter Class:" + str(adapter_instance.__class__))
            raise LaunchException("Developer Exception!!")

        ### Store Operation entity.
        algo = adapter_instance.stored_adapter
        algo_category = dao.get_category_by_id(algo.fk_category)

        operations = self.prepare_operations(current_user.id, project, algo, algo_category,
                                             {}, visible, view_model=model_view, **kwargs)[0]

        if isinstance(adapter_instance, ABCSynchronous):
            if len(operations) > 1:
                raise LaunchException("Synchronous operations are not supporting ranges!")
            if len(operations) < 1:
                self.logger.warning("No operation was defined")
                raise LaunchException("Invalid empty Operation!!!")
            return self.initiate_prelaunch(operations[0], adapter_instance, **kwargs)
        else:
            return self._send_to_cluster(operations, adapter_instance, current_user.username)

    @staticmethod
    def _prepare_metadata(initial_metadata, algo_category, operation_group, submit_data):
        """
        Gather metadata from submitted fields and current to be execute algorithm.
        Will populate STATE, GROUP in metadata
        """
        metadata = copy(initial_metadata)

        user_group = None
        if DataTypeMetaData.KEY_OPERATION_TAG in submit_data:
            user_group = submit_data[DataTypeMetaData.KEY_OPERATION_TAG]

        if operation_group is not None:
            metadata[DataTypeMetaData.KEY_OPERATION_TAG] = operation_group.name

        if DataTypeMetaData.KEY_TAG_1 in submit_data:
            metadata[DataTypeMetaData.KEY_TAG_1] = submit_data[DataTypeMetaData.KEY_TAG_1]

        metadata[DataTypeMetaData.KEY_STATE] = algo_category.defaultdatastate

        return metadata, user_group

    @staticmethod
    def _read_set(values):
        """ Parse a committed UI possible list of values, into a set converted into string."""
        if isinstance(values, list):
            set_values = []
            values_str = ""
            for val in values:
                if val not in set_values:
                    set_values.append(val)
                    values_str = values_str + " " + str(val)
            values = values_str
        return str(values).strip()

    def group_operation_launch(self, user_id, project, algorithm_id, category_id, existing_dt_group=None, **kwargs):
        """
        Create and prepare the launch of a group of operations.
        """
        category = dao.get_category_by_id(category_id)
        algorithm = dao.get_algorithm_by_id(algorithm_id)
        ops, _ = self.prepare_operations(user_id, project, algorithm, category, {},
                                         existing_dt_group=existing_dt_group, **kwargs)
        for operation in ops:
            self.launch_operation(operation.id, True)

    def _prepare_metric_operation(self, sim_operation):
        # type: (Operation) -> None
        metric_algo = dao.get_algorithm_by_module(TimeseriesMetricsAdapter.__module__,
                                                  TimeseriesMetricsAdapter.__name__)

        time_series_index = dao.get_generic_entity(TimeSeriesIndex, sim_operation.id, 'fk_from_operation')[0]
        ts_metrics_adapter_form = TimeseriesMetricsAdapterForm()
        ts_metrics_adapter_form.fill_from_trait(BaseTimeseriesMetricAlgorithm())
        ts_metrics_adapter_form.time_series.data = time_series_index.gid
        op_params = json.dumps(ts_metrics_adapter_form.get_dict())
        range_values = sim_operation.range_values
        metadata = {DataTypeMetaData.KEY_BURST: time_series_index.fk_parent_burst}
        metadata, user_group = self._prepare_metadata(metadata, metric_algo.algorithm_category, None, op_params)
        meta_str = json.dumps(metadata)

        parent_burst = dao.get_generic_entity(BurstConfiguration, time_series_index.fk_parent_burst, 'id')[0]
        metric_operation_group_id = parent_burst.metric_operation_group_id
        metric_operation = Operation(sim_operation.fk_launched_by, sim_operation.fk_launched_in, metric_algo.id,
                                     op_params,
                                     meta_str, op_group_id=metric_operation_group_id, range_values=range_values)
        metric_operation.visible = False
        operation = dao.store_entity(metric_operation)

        metrics_datatype_group = dao.get_generic_entity(DataTypeGroup, metric_operation_group_id, 'fk_operation_group')[
            0]
        if metrics_datatype_group.fk_from_operation is None:
            metrics_datatype_group.fk_from_operation = metric_operation.id

        return operation

    def prepare_operation(self, user_id, project_id, algorithm_id, category, view_model_gid, op_group, metadata,
                          ranges=None, visible=True):
        operation_parameters = json.dumps({'gid': view_model_gid})
        metadata, user_group = self._prepare_metadata(metadata, category, op_group, {})
        meta_str = json.dumps(metadata)

        op_group_id = None
        if op_group:
            op_group_id = op_group.id

        operation = Operation(user_id, project_id, algorithm_id, operation_parameters, op_group_id=op_group_id,
                              meta=meta_str, range_values=ranges)

        self.logger.debug("Saving Operation(userId=" + str(user_id) + ",projectId=" + str(project_id) + "," +
                          str(metadata) + ",algorithmId=" + str(algorithm_id) + ", ops_group= " + str(
            op_group_id) + ")")

        visible_operation = visible and category.display is False
        operation = dao.store_entity(operation)
        operation.visible = visible_operation

        return operation

    def prepare_operations(self, user_id, project, algorithm, category, metadata,
                           visible=True, existing_dt_group=None, view_model=None, **kwargs):
        """
        Do all the necessary preparations for storing an operation. If it's the case of a 
        range of values create an operation group and multiple operations for each possible
        instance from the range.
        :param metadata: Initial MetaData with potential Burst identification inside.
        """
        # TODO: fix group operations
        operations = []

        available_args, group = self._prepare_group(project.id, existing_dt_group, kwargs)
        if len(available_args) > TvbProfile.current.MAX_RANGE_NUMBER:
            raise LaunchException("Too big range specified. You should limit the"
                                  " resulting operations to %d" % TvbProfile.current.MAX_RANGE_NUMBER)
        else:
            self.logger.debug("Launching a range with %d operations..." % len(available_args))
        group_id = None
        if group is not None:
            group_id = group.id
        metadata, user_group = self._prepare_metadata(metadata, category, group, kwargs)

        self.logger.debug("Saving Operation(userId=" + str(user_id) + ",projectId=" + str(project.id) + "," +
                          str(metadata) + ",algorithmId=" + str(algorithm.id) + ", ops_group= " + str(group_id) + ")")

        visible_operation = visible and category.display is False
        meta_str = json.dumps(metadata)
        for (one_set_of_args, range_vals) in available_args:
            range_values = json.dumps(range_vals) if range_vals else None
            operation = Operation(user_id, project.id, algorithm.id,
                                  json.dumps({'gid': view_model.gid.hex}), meta_str,
                                  op_group_id=group_id, user_group=user_group, range_values=range_values)
            operation.visible = visible_operation
            operations.append(operation)
        operations = dao.store_entities(operations)

        if group is not None:
            burst_id = None
            if DataTypeMetaData.KEY_BURST in metadata:
                burst_id = metadata[DataTypeMetaData.KEY_BURST]
            if existing_dt_group is None:
                datatype_group = DataTypeGroup(group, operation_id=operations[0].id, fk_parent_burst=burst_id,
                                               state=metadata[DataTypeMetaData.KEY_STATE])
                dao.store_entity(datatype_group)
            else:
                # Reset count
                existing_dt_group.count_results = None
                dao.store_entity(existing_dt_group)

        for operation in operations:
            storage_path = FilesHelper().get_project_folder(project, str(operation.id))
            h5_path = h5.path_for(storage_path, ViewModelH5, view_model.gid)
            h5_file = ViewModelH5(h5_path, view_model)
            h5_file.store(view_model)
            h5_file.close()

        return operations, group

    def load_view_model(self, adapter_instance, operation):
        storage_path = self.file_helper.get_project_folder(operation.project, str(operation.id))
        input_gid = json.loads(operation.parameters)['gid']
        # TODO: review location, storage_path, op params deserialization
        if isinstance(adapter_instance, SimulatorAdapter):
            view_model = SimulatorSerializer().deserialize_simulator(input_gid, storage_path)
        else:
            view_model_class = adapter_instance.get_view_model_class()
            view_model = view_model_class()
            h5_path = h5.path_for(storage_path, ViewModelH5, input_gid)
            h5_file = ViewModelH5(h5_path, view_model)
            h5_file.load_into(view_model)
        return view_model

    def initiate_prelaunch(self, operation, adapter_instance, **kwargs):
        """
        Public method.
        This should be the common point in calling an adapter- method.
        """
        result_msg = ""
        temp_files = []
        try:
            unique_id = None
            if self.ATT_UID in kwargs:
                unique_id = kwargs[self.ATT_UID]

            operation = dao.get_operation_by_id(operation.id)  # Load Lazy fields

            disk_space_per_user = TvbProfile.current.MAX_DISK_SPACE
            pending_op_disk_space = dao.compute_disk_size_for_started_ops(operation.fk_launched_by)
            user_disk_space = dao.compute_user_generated_disk_size(operation.fk_launched_by)  # From kB to Bytes
            available_space = disk_space_per_user - pending_op_disk_space - user_disk_space

            view_model = self.load_view_model(adapter_instance, operation)
            result_msg, nr_datatypes = adapter_instance._prelaunch(operation, unique_id, available_space, view_model)
            operation = dao.get_operation_by_id(operation.id)
            ## Update DB stored kwargs for search purposes, to contain only valuable params (no unselected options)
            operation.parameters = json.dumps(kwargs)
            operation.mark_complete(STATUS_FINISHED)
            if nr_datatypes > 0:
                #### Write operation meta-XML only if some result are returned
                self.file_helper.write_operation_metadata(operation)
            dao.store_entity(operation)
            adapter_form = adapter_instance.get_form()
            try:
                temp_files = adapter_form.temporary_files
            except AttributeError:
                pass

            self._remove_files(temp_files)

        except zipfile.BadZipfile as excep:
            msg = "The uploaded file is not a valid ZIP!"
            self._handle_exception(excep, temp_files, msg, operation)
        except TVBException as excep:
            self._handle_exception(excep, temp_files, excep.message, operation)
        except MemoryError:
            msg = ("Could not execute operation because there is not enough free memory." +
                   " Please adjust operation parameters and re-launch it.")
            self._handle_exception(Exception(msg), temp_files, msg, operation)
        except Exception as excep1:
            msg = "Could not launch Operation with the given input data!"
            self._handle_exception(excep1, temp_files, msg, operation)

        if operation.fk_operation_group and 'SimulatorAdapter' in operation.algorithm.classname:
            next_op = self._prepare_metric_operation(operation)
            self.launch_operation(next_op.id)
        return result_msg

    def _send_to_cluster(self, operations, adapter_instance, current_username="unknown"):
        """ Initiate operation on cluster"""
        for operation in operations:
            try:
                BACKEND_CLIENT.execute(str(operation.id), current_username, adapter_instance)
            except Exception as excep:
                self._handle_exception(excep, {}, "Could not start operation!", operation)

        return operations

    def launch_operation(self, operation_id, send_to_cluster=False, adapter_instance=None):
        """
        Method exposed for Burst-Workflow related calls.
        It is used for cascading operation in the same workflow.
        """
        if operation_id is not None:
            operation = dao.get_operation_by_id(operation_id)
            if adapter_instance is None:
                algorithm = operation.algorithm
                adapter_instance = ABCAdapter.build_adapter(algorithm)

            if send_to_cluster:
                self._send_to_cluster([operation], adapter_instance, operation.user.username)
            else:
                self.initiate_prelaunch(operation, adapter_instance)

    def _handle_exception(self, exception, temp_files, message, operation=None):
        """
        Common way to treat exceptions:
            - remove temporary files, if any
            - set status ERROR on current operation (if any)
            - log exception
        """
        self.logger.exception(message)
        if operation is not None:
            BurstService().persist_operation_state(operation, STATUS_ERROR, str(exception))
        self._remove_files(temp_files)
        exception.message = message
        raise exception.with_traceback(
            sys.exc_info()[2])  # when rethrowing in python this is required to preserve the stack trace

    def _remove_files(self, file_list):
        """
        Remove any files that exist in the file_dictionary. 
        Currently used to delete temporary files created during an operation.
        """
        for pth in file_list:
            pth = str(pth)
            try:
                if os.path.exists(pth) and os.path.isfile(pth):
                    os.remove(pth)
                    self.logger.debug("We no longer need file:" + pth + " => deleted")
                else:
                    self.logger.warning("Trying to remove not existent file:" + pth)
            except OSError:
                self.logger.exception("Could not cleanup file!")

    @staticmethod
    def _range_name(range_no):
        return PARAM_RANGE_PREFIX + str(range_no)

    def _prepare_group(self, project_id, existing_dt_group, kwargs):
        """
        Create and store OperationGroup entity, or return None
        """
        # Standard ranges as accepted from UI
        range1_values = self.get_range_values(kwargs, self._range_name(1))
        range2_values = self.get_range_values(kwargs, self._range_name(2))
        available_args = self.__expand_arguments([(kwargs, None)], range1_values, self._range_name(1))
        available_args = self.__expand_arguments(available_args, range2_values, self._range_name(2))
        is_group = False
        ranges = []
        if self._range_name(1) in kwargs and range1_values is not None:
            is_group = True
            ranges.append(json.dumps((kwargs[self._range_name(1)], range1_values)))
        if self._range_name(2) in kwargs and range2_values is not None:
            is_group = True
            ranges.append(json.dumps((kwargs[self._range_name(2)], range2_values)))
        # Now for additional ranges which might be the case for the 'model exploration'
        last_range_idx = 3
        ranger_name = self._range_name(last_range_idx)
        while ranger_name in kwargs:
            values_for_range = self.get_range_values(kwargs, ranger_name)
            available_args = self.__expand_arguments(available_args, values_for_range, ranger_name)
            last_range_idx += 1
            ranger_name = self._range_name(last_range_idx)
        if last_range_idx > 3:
            ranges = []  # Since we only have 3 fields in db for this just hide it
        if not is_group:
            group = None
        elif existing_dt_group is None:
            group = OperationGroup(project_id=project_id, ranges=ranges)
            group = dao.store_entity(group)
        else:
            group = existing_dt_group.parent_operation_group

        return available_args, group

    def get_range_values(self, kwargs, ranger_name):
        """
        For the ranger given by ranger_name look in kwargs and return
        the array with all the possible values.
        """
        if ranger_name not in kwargs:
            return None
        if str(kwargs[ranger_name]) not in kwargs:
            return None

        range_values = []
        try:
            range_data = json.loads(str(kwargs[str(kwargs[ranger_name])]))
        except Exception:
            try:
                range_data = [x.strip() for x in str(kwargs[str(kwargs[ranger_name])]).split(',') if len(x.strip()) > 0]
                return range_data
            except Exception:
                self.logger.exception("Could not launch operation !")
                raise LaunchException("Could not launch with no data from:" + str(ranger_name))
        if type(range_data) in (list, tuple):
            return range_data

        if (constants.ATT_MINVALUE in range_data) and (constants.ATT_MAXVALUE in range_data):
            lo_val = float(range_data[constants.ATT_MINVALUE])
            hi_val = float(range_data[constants.ATT_MAXVALUE])
            step = float(range_data[constants.ATT_STEP])
            range_values = list(Range(lo=lo_val, hi=hi_val, step=step).to_array())  # , mode=Range.MODE_INCLUDE_BOTH))

        else:
            for possible_value in range_data:
                if range_data[possible_value]:
                    range_values.append(possible_value)
        return range_values

    @staticmethod
    def __expand_arguments(arguments_list, range_values, range_title):
        """
        Parse the arguments submitted from UI (flatten form) 
        If any ranger is found, return a list of arguments for all possible operations.
        """
        if range_values is None:
            return arguments_list
        result = []
        for value in range_values:
            for args, range_ in arguments_list:
                kw_new = copy(args)
                range_new = copy(range_)
                kw_new[kw_new[range_title]] = value
                if range_new is None:
                    range_new = {}
                range_new[kw_new[range_title]] = value
                del kw_new[range_title]
                result.append((kw_new, range_new))
        return result

    ##########################################################################################
    ######## Methods related to stopping and restarting operations start here ################
    ##########################################################################################

    def stop_operation(self, operation_id):
        """
        Stop the operation given by the operation id.
        """
        return BACKEND_CLIENT.stop_operation(int(operation_id))



import os
import tempfile
from uuid import UUID
import tvb_data
from tvb.adapters.datatypes.db.connectivity import ConnectivityIndex
from tvb.adapters.simulator.simulator_adapter import SimulatorAdapterModel
from tvb.adapters.uploaders.sensors_importer import SensorsImporterModel
from tvb.adapters.visualizers.sensors import SensorsViewerModel
from tvb.core.neocom import h5
from tvb.interfaces.rest.client.operation.operation_api import OperationApi
from tvb.interfaces.rest.client.simulator.simulation_api import SimulationApi
from tvb.tests.framework.core.base_testcase import TransactionalTestCase
from tvb.tests.framework.core.factory import TestFactory
import tvb_data.sensors as demo_data


class TestRestService(TransactionalTestCase):
    # Following tests are intended as examples and work when the server is started

    def transactional_setup_method(self):
        self.base_url = "http://127.0.0.1:9090/api"
        self.temp_folder = tempfile.gettempdir()
        self.test_user = TestFactory.create_user('Rest_User')
        self.test_project = TestFactory.create_project(self.test_user, 'Rest_Project')

    def test_launch_operation(self):
        operation_api = OperationApi(self.base_url)

        project_gid = "651a7f9c-3159-11ea-ada0-3464a92a26e5"
        algorithm_module = "tvb.adapters.uploaders.sensors_importer"
        algorithm_classname = "SensorsImporter"
        view_model = SensorsViewerModel()

        meg_file_path = os.path.join(os.path.dirname(demo_data.__file__), 'meg_151.txt.bz2')
        meg_sensors_index = TestFactory.import_sensors(self.test_user, self.test_project, meg_file_path,
                                                       SensorsImporterModel.OPTIONS['MEG Sensors'])
        meg_sensors = h5.load_from_index(meg_sensors_index)
        view_model.sensors = meg_sensors.gid

        operation_api.launch_operation(project_gid, algorithm_module, algorithm_classname, view_model,
                                       self.temp_folder)

    def test_fire_simulation(self):
        simulation_api = SimulationApi(self.base_url)

        session_stored_simulator = SimulatorAdapterModel()

        zip_path = os.path.join(os.path.dirname(tvb_data.__file__), 'connectivity', 'connectivity_76.zip')
        TestFactory.import_zip_connectivity(self.test_user, self.test_project, zip_path, "John")
        connectivity = TestFactory.get_entity(self.test_project, ConnectivityIndex)
        session_stored_simulator.connectivity = UUID(connectivity.gid)

        simulation_api.fire_simulation(self.test_project.gid, session_stored_simulator, None, self.temp_folder)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
import importlib
import os
import uuid
from tvb.core.entities.file.simulator.h5_factory import config_h5_factory
from tvb.core.neotraits.h5 import H5File
from tvb.core.neocom import h5


class SimulatorConfigurationH5(H5File):
    @staticmethod
    def get_full_class_name(class_entity):
        return class_entity.__module__ + '.' + class_entity.__name__

    def store_config_as_reference(self, config):
        gid = uuid.uuid4()

        config_h5_class = config_h5_factory(type(config))
        config_path = h5.path_for(os.path.dirname(self.path), config_h5_class, gid)

        with config_h5_class(config_path) as config_h5:
            config_h5.store(config)
            config_h5.gid.store(gid)
            config_h5.type.store(self.get_full_class_name(type(config)))

        return gid

    def get_reference_path(self, gid):
        dir_loader = h5.DirLoader(os.path.dirname(self.path), h5.REGISTRY)
        config_filename = dir_loader.find_file_name(gid)
        config_path = os.path.join(dir_loader.base_dir, config_filename)
        return config_path

    def load_from_reference(self, gid):
        config_path = self.get_reference_path(gid)

        config_h5 = H5File.from_file(config_path)

        config_type = config_h5.type.load()
        package, cls_name = config_type.rsplit('.', 1)
        module = importlib.import_module(package)
        config_class = getattr(module, cls_name)

        config_instance = config_class()
        config_h5.load_into(config_instance)
        config_h5.close()

        return config_instance



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Define DB changes for each release.

.. moduleauthor:: X Y <x.y@codemart.ro>

EXAMPLE FILE:

from sqlalchemy import Column, String
from migrate.changeset.schema import create_column, drop_column
from tvb.core.entities import model

COL_1 = Column('orientations_path', String)
COL_2 = Column('areas_path', String)

COL_3 = Column("local_connectivity_path", String)

COL_4 = Column("uid", String)
meta = model.Base.metadata
    

def upgrade(migrate_engine):
    "
    Upgrade operations go here. 
    Don't create your own engine; bind migrate_engine to your metadata.
    "
    meta.bind = migrate_engine
    table1 = meta.tables['DATA_connectivity']
    create_column(COL_1, table1)
    create_column(COL_2, table1)

    table2 = meta.tables['DATA_measuredcortex']
    create_column(COL_3, table2)
    
    table3 = meta.tables['DATA_TYPES']
    create_column(COL_4, table3)
    

def downgrade(migrate_engine):
    "
    Operations to reverse the above upgrade go here.
    "
    meta.bind = migrate_engine
    table1 = meta.tables['DATA_connectivity']
    table2 = meta.tables['DATA_measuredcortex']
    drop_column(COL_1, table1)
    drop_column(COL_2, table1)
    drop_column(COL_3, table2)
    
"""    
    
    


import uuid
from tvb.adapters.simulator.simulator_adapter import SimulatorAdapterModel
from tvb.core.adapters.abcadapter import ABCAdapter
from tvb.core.entities.file.simulator import h5_factory
from tvb.core.entities.file.simulator.cortex_h5 import CortexH5
from tvb.core.entities.file.simulator.simulator_h5 import SimulatorH5
from tvb.core.entities.storage import dao
from tvb.core.neocom import h5
from tvb.datatypes.sensors import SensorsEEG, SensorsMEG, SensorsInternal
from tvb.simulator.monitors import Projection, EEG, MEG, iEEG


class SimulatorSerializer(object):

    @staticmethod
    def serialize_simulator(simulator, simulator_gid, simulation_state_gid, storage_path):
        simulator_path = h5.path_for(storage_path, SimulatorH5, simulator_gid)

        with SimulatorH5(simulator_path) as simulator_h5:
            simulator_h5.gid.store(uuid.UUID(simulator_gid))
            simulator_h5.store(simulator)
            simulator_h5.connectivity.store(simulator.connectivity)
            if simulator.stimulus:
                simulator_h5.stimulus.store(simulator.stimulus)
            if simulation_state_gid:
                simulator_h5.simulation_state.store(uuid.UUID(simulation_state_gid))

        return simulator_gid

    @staticmethod
    def deserialize_simulator(simulator_gid, storage_path):
        simulator_in_path = h5.path_for(storage_path, SimulatorH5, simulator_gid)
        simulator_in = SimulatorAdapterModel()

        with SimulatorH5(simulator_in_path) as simulator_in_h5:
            simulator_in_h5.load_into(simulator_in)
            simulator_in.connectivity = simulator_in_h5.connectivity.load()
            simulator_in.stimulus = simulator_in_h5.stimulus.load()
            simulator_in.history_gid = simulator_in_h5.simulation_state.load()

        if isinstance(simulator_in.monitors[0], Projection):
            # TODO: simplify this part
            with SimulatorH5(simulator_in_path) as simulator_in_h5:
                monitor_h5_path = simulator_in_h5.get_reference_path(simulator_in.monitors[0].gid)

            monitor_h5_class = h5_factory.monitor_h5_factory(type(simulator_in.monitors[0]))

            with monitor_h5_class(monitor_h5_path) as monitor_h5:
                sensors_gid = monitor_h5.sensors.load()
                region_mapping_gid = monitor_h5.region_mapping.load()

            sensors_index = ABCAdapter.load_entity_by_gid(sensors_gid.hex)
            sensors = h5.load_from_index(sensors_index)

            if isinstance(simulator_in.monitors[0], EEG):
                sensors = SensorsEEG.build_sensors_subclass(sensors)
            elif isinstance(simulator_in.monitors[0], MEG):
                sensors = SensorsMEG.build_sensors_subclass(sensors)
            elif isinstance(simulator_in.monitors[0], iEEG):
                sensors = SensorsInternal.build_sensors_subclass(sensors)

            simulator_in.monitors[0].sensors = sensors
            region_mapping_index = ABCAdapter.load_entity_by_gid(region_mapping_gid.hex)
            region_mapping = h5.load_from_index(region_mapping_index)
            simulator_in.monitors[0].region_mapping = region_mapping

        if simulator_in.surface:
            cortex_path = h5.path_for(storage_path, CortexH5, simulator_in.surface.gid)
            with CortexH5(cortex_path) as cortex_h5:
                simulator_in.surface.local_connectivity = cortex_h5.local_connectivity.load()
                simulator_in.surface.region_mapping_data = cortex_h5.region_mapping_data.load()
                rm_index = dao.get_datatype_by_gid(simulator_in.surface.region_mapping_data.hex)
                simulator_in.surface.surface_gid = uuid.UUID(rm_index.surface_gid)

        return simulator_in



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Ionel Ortelecan <ionel.ortelecan@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>

Data for Parameter Space Exploration view will be defined here.
The purpose of this entities is to be used in Jinja2 UI, or for populating visualizer.
"""

import json
import math
import six
from tvb.basic.config.utils import EnhancedDictionary
from tvb.core.entities.model.model_operation import STATUS_FINISHED


class ContextDiscretePSE(EnhancedDictionary):
    """
    Entity used for filling a PSE visualizer.
    """
    KEY_GID = "Gid"
    KEY_NODE_TYPE = "dataType"
    KEY_OPERATION_ID = "operationId"
    KEY_TOOLTIP = "tooltip"
    LINE_SEPARATOR = "<br/>"


    def __init__(self, datatype_group_gid, color_metric, size_metric, back_page):

        super(ContextDiscretePSE, self).__init__()

        self.datatype_group_gid = datatype_group_gid
        self.min_color = float('inf')
        self.max_color = - float('inf')
        self.min_shape_size = float('inf')
        self.max_shape_size = - float('inf')
        self.has_started_ops = False
        self.status = 'started'
        self.title_x, self.title_y = "", ""
        self.values_x, self.labels_x, self.values_y, self.labels_y = [], [], [], []
        self.series_array, self.available_metrics = [], []
        self.datatypes_dict = {}
        self.d3_data = {}
        self.color_metric = color_metric
        self.size_metric = size_metric
        self.pse_back_page = back_page


    def setRanges(self, title_x, values_x, labels_x, title_y, values_y, labels_y, only_numbers):
        self.title_x = title_x
        self.values_x = values_x
        self.labels_x = labels_x

        self.title_y = title_y
        self.values_y = values_y
        self.labels_y = labels_y
        self.only_numbers = only_numbers


    def prepare_individual_jsons(self):
        """
        Apply JSON.dumps on all attributes which can not be passes as they are towards UI.
        """
        self.labels_x = json.dumps(self.labels_x)
        self.labels_y = json.dumps(self.labels_y)
        self.values_x = json.dumps(self.values_x)
        self.values_y = json.dumps(self.values_y)
        self.d3_data = json.dumps(self.d3_data)
        self.has_started_ops = json.dumps(self.has_started_ops)


    def prepare_full_json(self):
        """
        Apply JSON.dumps on full dictionary.
        """
        return json.dumps(self)


    def build_node_info(self, operation, datatype):
        """
        Build a dictionary with all the required information to be displayed for a given node.
        """
        node_info = {}
        if operation.status == STATUS_FINISHED and datatype is not None:
            ### Prepare attributes to be able to show overlay and launch further analysis.
            node_info[self.KEY_GID] = datatype.gid
            node_info[self.KEY_NODE_TYPE] = datatype.type
            node_info[self.KEY_OPERATION_ID] = operation.id
            ### Prepare tooltip for quick display.
            datatype_tooltip = str("Operation id: " + str(operation.id) + self.LINE_SEPARATOR +
                                   "Datatype gid: " + str(datatype.gid) + self.LINE_SEPARATOR +
                                   "Datatype type: " + str(datatype.type) + self.LINE_SEPARATOR +
                                   "Datatype subject: " + str(datatype.subject) + self.LINE_SEPARATOR +
                                   "Datatype invalid: " + str(datatype.invalid))
            ### Add scientific report to the quick details.
            if hasattr(datatype, 'summary_info') and datatype.summary_info is not None:
                for key, value in six.iteritems(datatype.summary_info):
                    datatype_tooltip = datatype_tooltip + self.LINE_SEPARATOR + str(key) + ": " + str(value)
            node_info[self.KEY_TOOLTIP] = datatype_tooltip
        else:
            tooltip = "No result available. Operation is in status: %s" % operation.status.split('-')[1]
            node_info[self.KEY_TOOLTIP] = tooltip
        return node_info


    def prepare_metrics_datatype(self, measures, datatype):
        """
        Update attribute self.datatypes_dict with metric values for this DataType.
        """
        dt_info = {}
        if measures is not None and len(measures) > 0:
            measure = measures[0]
            metrics = json.loads(measure.metrics)
            self.available_metrics = list(metrics)

            # As default we have the first two metrics available is no metrics are passed from the UI
            if self.color_metric is None and self.size_metric is None:
                if len(self.available_metrics) >= 1:
                    self.color_metric = self.available_metrics[0]
                if len(self.available_metrics) >= 2:
                    self.size_metric = self.available_metrics[1]

            if self.color_metric is not None:
                color_value = metrics[self.color_metric]
                if color_value < self.min_color:
                    self.min_color = color_value
                if color_value > self.max_color:
                    self.max_color = color_value
                dt_info[self.color_metric] = color_value

            if self.size_metric is not None:
                size_value = metrics[self.size_metric]
                if size_value < self.min_shape_size:
                    self.min_shape_size = size_value
                if size_value > self.max_shape_size:
                    self.max_shape_size = size_value
                dt_info[self.size_metric] = size_value
        self.datatypes_dict[datatype.gid] = dt_info


    def fill_object(self, final_dict):
        """ Populate current entity with attributes required for visualizer"""
        all_series = []
        if self.only_numbers:
            self.values_x = list(final_dict)

            y_values_set = set()
            for dict_ in final_dict.values():
                y_values_set = y_values_set.union(list(dict_))
            self.values_y = [a for a in y_values_set]

            self.values_x.sort()
            self.values_y.sort()
            self.labels_x = self.values_x
            self.labels_y = self.values_y

        for key_1 in final_dict.keys():
            for key_2 in final_dict[key_1].keys():
                datatype_gid = None
                current = final_dict[key_1][key_2]
                if self.KEY_GID in current:
                    # This means the operation was finished
                    datatype_gid = current[self.KEY_GID]

                parameter_coords = '{"x":' + str(key_1) + ', "y":' + str(key_2) + '}'
                color_weight, shape_type_1 = self.__get_color_weight(self.datatypes_dict, datatype_gid,
                                                                     self.color_metric)
                if (shape_type_1 is not None) and (datatype_gid is not None):
                    current[self.KEY_TOOLTIP] += self.LINE_SEPARATOR + " Color metric has NaN values"
                shape_size, shape_type_2 = self.__get_node_size(self.datatypes_dict, datatype_gid, len(self.labels_x),
                                                                len(self.labels_y), self.size_metric)
                if (shape_type_2 is not None) and (datatype_gid is not None):
                    current[self.KEY_TOOLTIP] += self.LINE_SEPARATOR + " Size metric has NaN values"
                # If either of the shape_types is not none use that
                shape_type = shape_type_1 or shape_type_2
                series = self.__get_node_json(shape_type, shape_size, parameter_coords)
                current['color_weight'] = color_weight
                all_series.append(series)

        self.d3_data = final_dict
        self.series_array = self.__build_series_json(all_series)
        self.status = 'started' if self.has_started_ops else 'finished'


    @staticmethod
    def __build_series_json(list_of_series):
        """ Given a list with all the data points, build the final FLOT JSON. """
        final_json = "["
        for i, value in enumerate(list_of_series):
            if i:
                final_json += ","
            final_json += value
        final_json += "]"
        return final_json


    @staticmethod
    def __get_node_json(symbol, radius, coords):
        """
        For each data point entry, build the FLOT specific JSON.
        """
        series = '{"points": {'
        if symbol is not None:
            series += '"symbol": "' + symbol + '", '
        series += '"radius": ' + str(radius) + '}, '
        series += '"coords": ' + coords + '}'
        return series


    @staticmethod
    def __get_color_weight(datatype_indexes, datatype_gid, metric):
        """
        Returns the color weight of the shape used for representing the dataType with id equal to 'datatype_id'.

        :param: datatype_indexes -  a dictionary which contains as keys a dataType GID and as values
                            the index corresponding to the dataType into the list of
                            dataTypes resulted from the same operation group
        :param: datatype_gid - It should exists into the 'datatype_indexes' dictionary.
        :param: metric - current metric key.
        """
        if datatype_gid is None:
            return 0, "cross"
        node_info = datatype_indexes[datatype_gid]
        valid_metric = True

        if metric in node_info:
            try:
                if math.isnan(float(node_info[metric])) or math.isinf(float(node_info[metric])):
                    valid_metric = False
            except ValueError:
                valid_metric = False
            if valid_metric:
                return node_info[metric], None
            else:
                return 0, "cross"
        return 0, None


    def __get_node_size(self, datatype_indexes, datatype_gid, range1_length, range2_length, metric):
        """
        Computes the size of the shape used for representing the dataType with GID given.

        :param: datatype_indexes -  a dictionary which contains as keys a daTatype GID and as values
                            the index corresponding to the dataType into the list of
                            dataTypes resulted from the same operation group
        :param: datatype_gid - It should exists into the 'datatype_indexes' dict.
        
        """
        min_size, max_size = self.__get_boundaries(range1_length, range2_length)
        if datatype_gid is None:
            return max_size / 2.0, "cross"
        node_info = datatype_indexes[datatype_gid]

        if metric in node_info:
            valid_metric = True
            try:
                if math.isnan(float(node_info[metric])) or math.isinf(float(node_info[metric])):
                    valid_metric = False
            except ValueError:
                valid_metric = False

            if valid_metric:
                shape_weight = node_info[metric]
                values_range = self.max_shape_size - self.min_shape_size
                shape_range = max_size - min_size
                if values_range != 0:
                    return min_size + (shape_weight - self.min_shape_size) / float(values_range) * shape_range, None
                else:
                    return min_size, None
            else:
                return max_size / 2.0, "cross"
        return max_size / 2.0, None


    @staticmethod
    def __get_boundaries(range1_length, range2_length):
        """
        Returns the MIN and the max values of the interval from
        which may be set the size of a certain shape.
        """
        # the arrays 'intervals' and 'values' should have the same size
        intervals = [0, 3, 5, 8, 10, 20, 30, 40, 50, 60, 90, 110, 120]
        values = [(10, 50), (10, 40), (10, 33), (8, 25), (5, 15), (4, 10),
                  (3, 8), (2, 6), (2, 5), (1, 4), (1, 3), (1, 2), (1, 2)]

        max_length = max([range1_length, range2_length])
        if max_length <= intervals[0]:
            return values[0]
        elif max_length >= intervals[-1]:
            return values[-1]
        else:
            for i, interval in enumerate(intervals):
                if max_length <= interval:
                    return values[i - 1]



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Ionel Ortelecan <ionel.ortelecan@codemart.ro>
"""

from tvb.core.entities.storage import dao 

MAX_SHAPE_SIZE = 50
MIN_SHAPE_SIZE = 10

NODE_DATATYPE_TYPE = "datatype"
NODE_OPERATION_TYPE = "operation"
NODE_OPERATION_GROUP_TYPE = "operationGroup"



class NodeData(object):
    """
    Contains the meta-data that will be set on each GRAPH node.
    """

    shape_size = None
    shape_color = None
    shape_type = None
    node_type = None
    node_entity_id = None
    node_subtitle = None


    def __init__(self, shape_size, shape_color, shape_type, node_type, node_entity_id, node_subtitle):
        self.shape_size = shape_size
        self.shape_color = shape_color
        self.shape_type = shape_type
        self.node_type = node_type
        self.node_entity_id = node_entity_id
        self.node_subtitle = node_subtitle


    def to_json(self):
        """
        Returns the JSON-ready representation of this NodeData instance.
        """
        instance_json = {"$dim": self.shape_size,
                         "$color": self.shape_color,
                         "$type": self.shape_type,
                         "node_type": self.node_type,
                         "node_entity_id": self.node_entity_id,
                         "node_subtitle": self.node_subtitle}
        return instance_json


    @staticmethod
    def build_node_for_datatype(datatype_id, node_subtitle, shape_size=MAX_SHAPE_SIZE, is_group=False):

        shape = "circlesGroup" if is_group else "circle"
        return NodeData(shape_size, "#83548B", shape, NODE_DATATYPE_TYPE, datatype_id, node_subtitle)


    @staticmethod
    def build_node_for_operation(operation, group_id=None):

        if group_id:
            entity_id = group_id
            node_type = NODE_OPERATION_GROUP_TYPE
            shape = "squaresGroup"
        else:
            entity_id = operation.id
            node_type = NODE_OPERATION_TYPE
            shape = "square"

        return NodeData(MAX_SHAPE_SIZE, "#660033", shape, node_type, entity_id, str(operation.start_date))



class NodeStructure(object):
    """
    Define the full structure of a graph NODE (including meta-data of type NodeData and node Adjiacences)
    """

    node_gid = None
    name = None
    data = None
    adjacencies = []
    selected = False


    def __init__(self, node_gid, node_name):
        self.node_gid = node_gid
        self.name = node_name


    def to_json(self):
        """
        Returns the JSON-ready representation of this NodeStructure instance.
        """
        instance_json = {"id": self.node_gid,
                         "name": self.name,
                         "data": self.data.to_json() if self.data is not None else {},
                         "adjacencies": [{"nodeFrom": self.node_gid, "nodeTo": adj,
                                          "data": {}} for adj in self.adjacencies]}
        return instance_json


    @staticmethod
    def build_structure_for_datatype(datatype_gid):

        datatype = dao.get_datatype_by_gid(datatype_gid)
        is_group = dao.is_datatype_group(datatype_gid)

        structure = NodeStructure(datatype_gid, datatype.type)
        structure.data = NodeData.build_node_for_datatype(datatype.id, datatype.display_name, is_group=is_group)
        return structure


    @staticmethod
    def build_structure_for_operation(operation):

        algo = dao.get_algorithm_by_id(operation.fk_from_algo)

        structure = NodeStructure(operation.gid, algo.displayname)
        structure.data = NodeData.build_node_for_operation(operation)
        return structure


    @staticmethod
    def build_structure_for_operation_group(operation_group_gid):

        group = dao.get_operationgroup_by_gid(operation_group_gid)
        operation = dao.get_operations_in_group(group.id, only_first_operation=True)
        algo = dao.get_algorithm_by_id(operation.fk_from_algo)

        structure = NodeStructure(operation_group_gid, algo.displayname)
        structure.data = NodeData.build_node_for_operation(operation, group.id)
        return structure


    @staticmethod
    def build_artificial_root_structure(adjacencies_nodes):

        root_structure = NodeStructure("fakeRootNode", "fakeRootNode")
        root_structure.data = NodeData.build_node_for_datatype("fakeRootNode", "Fake root",
                                                               shape_size=MAX_SHAPE_SIZE / 2)
        root_structure.adjacencies = adjacencies_nodes
        return root_structure



class GraphComponent():
    """
    Class used for representing a single component of the graph.
    One GraphComponent holds multiple lists of NodeStructure instances (for multiple layers).

    A GraphComponent will have the following structure::

        input_datatypes             operation_parent
            |                               |
            |                               |
            V                               V
        [operation_parent]          [output_datatypes]             * Currently Selected node
            |                               |
            |                               |
            V                               V
        output_datatypes            in_operations

    """
    input_datatypes = []
    operation_parent = []
    output_datatypes = []
    in_operations = []


    def __init__(self, dt_inputs, parent_op, dt_outputs, op_inputs):
        self.input_datatypes = dt_inputs
        self.operation_parent = parent_op
        self.output_datatypes = dt_outputs
        self.in_operations = op_inputs



class FullGraphStructure():
    """
    This class contains information for the entire graph to be displayed in UI.
    It holds a list of GraphComponent instances (e.g. multiple UPLOAD ops).
    """

    graph_components = []


    def __init__(self, components):
        self.graph_components = components
        self.fill_shape_size()
        self.fill_all_graph_adjiacences()


    def prepare_for_json(self):
        """
        Returns a list of NodeStructure instances to be serialized for browser-client rendering.
        """
        artificial_root_adj = []
        for component in self.graph_components:
            if len(component.input_datatypes):
                artificial_root_adj.extend(self._get_nodes_gids(component.input_datatypes))
            else:
                artificial_root_adj.extend(self._get_nodes_gids(component.operation_parent))

        result_to_serialize = [NodeStructure.build_artificial_root_structure(artificial_root_adj)]
        for component in self.graph_components:
            for level in [component.input_datatypes, component.operation_parent,
                          component.output_datatypes, component.in_operations]:
                for node_structure in level:
                    result_to_serialize.append(node_structure)

        return result_to_serialize


    def fill_all_graph_adjiacences(self):

        for branch in self.graph_components:
            parent_ops_gids = self._get_nodes_gids(branch.operation_parent)
            dt_outputs_gids = self._get_nodes_gids(branch.output_datatypes)
            op_inputs_gids = self._get_nodes_gids(branch.in_operations)

            self._set_nodes_adjacencies(branch.input_datatypes, parent_ops_gids)
            self._set_nodes_adjacencies(branch.operation_parent, dt_outputs_gids)
            self._set_nodes_adjacencies(branch.output_datatypes, op_inputs_gids, True)


    @staticmethod
    def _get_nodes_gids(list_of_nodes):
        return [node.node_gid for node in list_of_nodes]


    @staticmethod
    def _set_nodes_adjacencies(list_of_nodes, adjacencies, only_for_selected_node=False):
        """
        Sets adjacencies for a list of nodes.
        """
        if only_for_selected_node:
            for node in list_of_nodes:
                if node.selected:
                    node.adjacencies = adjacencies
                else:
                    node.adjacencies = []
        else:
            for node in list_of_nodes:
                node.adjacencies = adjacencies


    def fill_shape_size(self):
        """
        Sets the correct size for each node from this graph.
        """
        no_of_dt_inputs = 0
        no_of_parent_op = 0
        no_of_dt_outputs = 0
        no_of_op_inputs = 0

        for branch in self.graph_components:
            no_of_dt_inputs += len(branch.input_datatypes)
            no_of_parent_op += len(branch.operation_parent)
            no_of_dt_outputs += len(branch.output_datatypes)
            no_of_op_inputs += len(branch.in_operations)

        dt_input_size = self._compute_shape_size(no_of_dt_inputs)
        parent_op_size = self._compute_shape_size(no_of_parent_op)
        dt_output_size = self._compute_shape_size(no_of_dt_outputs)
        op_input = self._compute_shape_size(no_of_op_inputs)

        for branch in self.graph_components:
            self._set_nodes_size(branch.input_datatypes, dt_input_size)
            self._set_nodes_size(branch.operation_parent, parent_op_size)
            self._set_nodes_size(branch.output_datatypes, dt_output_size)
            self._set_nodes_size(branch.in_operations, op_input)


    @staticmethod
    def _set_nodes_size(list_of_nodes, shape_size):
        """
        Sets the size for each node from the given list of nodes.
        """
        if not len(list_of_nodes):
            return

        for node in list_of_nodes:
            if node.data is None:
                continue
            node.data.shape_size = shape_size


    @staticmethod
    def _compute_shape_size(no_of_elements):
        """
        no_of_elements - represents the number of nodes that are displayed on a certain level

        We consider that the canvas width is 1000px. With a size of 50 the max number
        of shapes that may be displayed into the canvas are 9 (=> 111px per shape).

        max_shape_size = 50 points <=> 111px
        min_shape_size = 10 points <=> 20px

        1 point <=> 2px => shape_size = (canvas_width) / [(no_of_nodes + 1) * 2]
        """

        if not no_of_elements:
            return MAX_SHAPE_SIZE

        shape_size = 1000 / ((no_of_elements + 1) * 2)
        if shape_size > MAX_SHAPE_SIZE:
            return MAX_SHAPE_SIZE
        elif shape_size < MIN_SHAPE_SIZE:
            return MIN_SHAPE_SIZE
        else:
            return shape_size





# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

import os
import json
import uuid
from collections import namedtuple
from datetime import datetime
import numpy
from tvb.core import utils
from tvb.core.entities.file.files_helper import FilesHelper
from tvb.core.entities.model.model_datatype import DataType
from tvb.core.entities.storage import dao
from tvb.core.entities.filters.chain import FilterChain
from tvb.core.neocom.h5 import REGISTRY
from tvb.basic.neotraits.ex import TraitError
from tvb.basic.neotraits.api import List, Attr
from tvb.core.neotraits.view_model import DataTypeGidAttr

# This setting is injected.
# The pattern might be confusing, but it is an interesting alternative to
# universal tvbprofile imports

jinja_env = None


def prepare_prefixed_name_for_field(prefix, name):
    return '{}_{}'.format(prefix, name)


class Field(object):
    template = None

    def __init__(self, form, name, disabled=False, required=False, label='', doc='', default=None):
        # type: (Form, str, bool, bool, str, str, object) -> None
        self.owner = form
        self.name = prepare_prefixed_name_for_field(form.prefix, name)
        self.disabled = disabled
        self.required = required
        self.label = label
        self.doc = doc
        self.label_classes = []
        if required:
            self.label_classes.append('field-mandatory')

        # keeps the deserialized data
        self.data = None
        # keeps user input, even if wrong, we have to redisplay it
        # todo
        self.unvalidated_data = default
        self.errors = []

    def fill_from_post(self, post_data):
        """ deserialize form a post dictionary """
        self.unvalidated_data = post_data.get(self.name)
        try:
            self._from_post()
        except ValueError as ex:
            self.errors.append(ex)

    def validate(self):
        """ validation besides the deserialization from post"""
        return not self.errors

    def _from_post(self):
        if self.required and self.unvalidated_data is None:
            raise ValueError('Field required')
        self.data = self.unvalidated_data

    @property
    def value(self):
        if str(self.data) == self.unvalidated_data:
            return self.data
        return self.data or self.unvalidated_data

    def __repr__(self):
        return '<{}>(name={})'.format(type(self).__name__, self.name)

    def __str__(self):
        return jinja_env.get_template(self.template).render(field=self)


class SimpleBoolField(Field):
    template = 'form_fields/bool_field.html'

    def _from_post(self):
        if self.unvalidated_data is None:
            self.data = False
        self.data = bool(self.unvalidated_data)


class SimpleStrField(Field):
    template = 'form_fields/str_field.html'

    def _from_post(self):
        if self.required and (self.unvalidated_data is None or self.unvalidated_data.strip() == ''):
            raise ValueError('Field required')
        self.data = self.unvalidated_data


class SimpleHiddenField(Field):
    template = 'form_fields/hidden_field.html'


class SimpleIntField(Field):
    template = 'form_fields/number_field.html'
    min = None
    max = None

    def _from_post(self):
        super(SimpleIntField, self)._from_post()
        if self.unvalidated_data is None or self.unvalidated_data.strip() == '':
            if self.required:
                raise ValueError('Field required')
            self.data = None
        else:
            self.data = int(self.unvalidated_data)


class SimpleFloatField(Field):
    template = 'form_fields/number_field.html'
    input_type = "number"
    min = None
    max = None
    step = 'any'

    def _from_post(self):
        super(SimpleFloatField, self)._from_post()
        if self.unvalidated_data is None or self.unvalidated_data.strip() == '':
            if self.required:
                raise ValueError('Field required')
            self.data = None
        else:
            self.data = float(self.unvalidated_data)


class SimpleSelectField(Field):
    template = 'form_fields/radio_field.html'
    missing_value = 'explicit-None-value'

    def __init__(self, choices, form, name=None, disabled=False, required=False, label='', doc='', default=None,
                 include_none=True):
        super(SimpleSelectField, self).__init__(form, name, disabled, required, label, doc, default)
        self.choices = choices
        self.include_none = include_none

    def options(self):
        """ to be used from template, assumes self.data is set """
        if not self.required and self.include_none:
            choice = None
            yield Option(
                id='{}_{}'.format(self.name, None),
                value=self.missing_value,
                label=str(choice).title(),
                checked=self.data is None
            )

        for i, choice in enumerate(self.choices.keys()):
            yield Option(
                id='{}_{}'.format(self.name, i),
                value=choice,
                label=str(choice).title(),
                checked=self.data == self.choices.get(choice)
            )

    def fill_from_post(self, post_data):
        super(SimpleSelectField, self).fill_from_post(post_data)
        self.data = self.choices.get(self.data)


class MultipleSelectField(SimpleSelectField):
    template = 'form_fields/checkbox_field.html'

    def fill_from_post(self, post_data):
        super(SimpleSelectField, self).fill_from_post(post_data)
        if self.data is not None:
            if not isinstance(self.data, list):
                self.data = [self.data]
            data = list()
            for choice in self.data:
                data.append(self.choices.get(choice))
            self.data = data


class SimpleArrayField(Field):
    template = 'form_fields/str_field.html'

    def __init__(self, form, name, dtype, disabled=False, required=False, label='', doc='', default=None):
        super(SimpleArrayField, self).__init__(form, name, disabled, required, label, doc, default)
        self.dtype = dtype

    def _from_post(self):
        if self.unvalidated_data is not None and isinstance(self.unvalidated_data, str):
            data = json.loads(self.unvalidated_data)
            self.data = numpy.array(data, dtype=self.dtype).tolist()
        elif self.unvalidated_data is not None and isinstance(self.unvalidated_data, list):
            self.data = self.unvalidated_data

    @property
    def value(self):
        if self.data is None:
            # todo: maybe we need to distinguish None from missing data
            # this None means self.data is missing, either not set or unset cause of validation error
            return self.unvalidated_data
        try:
            return json.dumps(self.data.tolist())
        except (TypeError, ValueError):
            return self.unvalidated_data


class DataTypeSelectField(Field):
    template = 'form_fields/datatype_select_field.html'
    missing_value = 'explicit-None-value'

    def __init__(self, datatype_index, form, name=None, disabled=False, required=False, label='', doc='',
                 conditions=None, draw_dynamic_conditions_buttons=True, dynamic_conditions=None, has_all_option=False):
        super(DataTypeSelectField, self).__init__(form, name, disabled, required, label, doc)
        self.datatype_index = datatype_index
        self.conditions = conditions
        self.draw_dynamic_conditions_buttons = draw_dynamic_conditions_buttons
        self.dynamic_conditions = dynamic_conditions
        self.has_all_option = has_all_option

    @property
    def get_dynamic_filters(self):
        return FilterChain().get_filters_for_type(self.datatype_index)

    def _get_values_from_db(self):
        all_conditions = FilterChain()
        all_conditions += self.conditions
        all_conditions += self.dynamic_conditions
        filtered_datatypes, count = dao.get_values_of_datatype(self.owner.project_id,
                                                               self.datatype_index,
                                                               all_conditions)
        return filtered_datatypes

    def options(self):
        if not self.owner.project_id:
            raise ValueError('A project_id is required in order to query the DB')

        filtered_datatypes = self._get_values_from_db()

        if not self.required:
            choice = None
            yield Option(
                id='{}_{}'.format(self.name, None),
                value=self.missing_value,
                label=str(choice).title(),
                checked=self.data is None
            )

        for i, datatype in enumerate(filtered_datatypes):
            yield Option(
                id='{}_{}'.format(self.name, i),
                value=datatype[2],
                label=self._prepare_display_name(datatype),
                checked=self.data == datatype[2]
            )

        if self.has_all_option:
            if not self.owner.draw_ranges:
                raise ValueError("The owner form should draw ranges inputs in order to support 'All' option")

            all_values = ''
            for fdt in filtered_datatypes:
                all_values += str(fdt[2]) + ','

            choice = "All"
            yield Option(
                id='{}_{}'.format(self.name, choice),
                value=all_values[:-1],
                label=choice,
                checked=self.data is choice
            )

    def get_dt_from_db(self):
        return dao.get_datatype_by_gid(self.data)

    def _prepare_display_name(self, value):
        """
        Populate meta-data fields for data_list (list of DataTypes).

        Private method, to be called recursively.
        It will receive a list of Attributes, and it will populate 'options'
        entry with data references from DB.
        """
        # Here we only populate with DB data, actual
        # XML check will be done after select and submit.
        entity_gid = value[2]
        actual_entity = dao.get_generic_entity(self.datatype_index, entity_gid, "gid")
        display_name = ''
        if actual_entity is not None and len(actual_entity) > 0 and isinstance(actual_entity[0], DataType):
            display_name = actual_entity[0].__class__.__name__
        display_name += ' - ' + (value[3] or "None ")
        if value[5]:
            display_name += ' - From: ' + str(value[5])
        else:
            display_name += utils.date2string(value[4])
        if value[6]:
            display_name += ' - ' + str(value[6])
        display_name += ' - ID:' + str(value[0])

        return display_name

    def _from_post(self):
        if self.unvalidated_data == self.missing_value:
            self.unvalidated_data = None

        if self.required and not self.unvalidated_data:
            raise ValueError('Field required')

        # TODO: ensure is in choices
        self.data = self.unvalidated_data


class TraitField(Field):
    # mhtodo: while this is consistent with the h5 api, it has the same problem
    #         it couples the system to traited attr declarations
    def __init__(self, trait_attribute, form, name=None, disabled=False):
        # type: (Attr, Form, str, bool) -> None
        self.trait_attribute = trait_attribute  # type: Attr
        name = name or trait_attribute.field_name
        label = trait_attribute.label or name

        super(TraitField, self).__init__(
            form,
            name,
            disabled,
            trait_attribute.required,
            label,
            trait_attribute.doc,
            trait_attribute.default
        )

    def from_trait(self, trait, f_name):
        self.data = getattr(trait, f_name)


TEMPORARY_PREFIX = ".tmp"


class TraitUploadField(TraitField):
    template = 'form_fields/upload_field.html'

    def __init__(self, traited_attribute, required_type, form, name, disabled=False):
        super(TraitUploadField, self).__init__(traited_attribute, form, name, disabled)
        self.required_type = required_type
        self.files_helper = FilesHelper()

    def fill_from_post(self, post_data):
        super(TraitUploadField, self).fill_from_post(post_data)

        if self.data.file is None:
            self.data = None
            return

        project = dao.get_project_by_id(self.owner.project_id)
        temporary_storage = self.files_helper.get_project_folder(project, self.files_helper.TEMP_FOLDER)

        file_name = None
        try:
            uq_name = utils.date2string(datetime.now(), True) + '_' + str(0)
            file_name = TEMPORARY_PREFIX + uq_name + '_' + self.data.filename
            file_name = os.path.join(temporary_storage, file_name)

            with open(file_name, 'wb') as file_obj:
                file_obj.write(self.data.file.read())
        except Exception as excep:
            # TODO: is this handled properly?
            self.files_helper.remove_files([file_name])
            excep.message = 'Could not continue: Invalid input files'
            raise excep

        if file_name:
            self.data = file_name
            self.owner.temporary_files.append(file_name)


class TraitDataTypeSelectField(TraitField):
    template = 'form_fields/datatype_select_field.html'
    missing_value = 'explicit-None-value'

    def __init__(self, trait_attribute, form, name=None, conditions=None, draw_dynamic_conditions_buttons=True,
                 dynamic_conditions=None, has_all_option=False):
        super(TraitDataTypeSelectField, self).__init__(trait_attribute, form, name)
        if issubclass(type(trait_attribute), DataTypeGidAttr):
            type_to_query = trait_attribute.linked_datatype
        else:
            type_to_query = trait_attribute.field_type

        self.datatype_index = REGISTRY.get_index_for_datatype(type_to_query)
        self.conditions = conditions
        self.draw_dynamic_conditions_buttons = draw_dynamic_conditions_buttons
        self.dynamic_conditions = dynamic_conditions
        self.has_all_option = has_all_option

    def from_trait(self, trait, f_name):
        if hasattr(trait, f_name):
            self.data = getattr(trait, f_name)

    @property
    def get_dynamic_filters(self):
        return FilterChain().get_filters_for_type(self.datatype_index)

    def _get_values_from_db(self):
        all_conditions = FilterChain()
        all_conditions += self.conditions
        all_conditions += self.dynamic_conditions
        filtered_datatypes, count = dao.get_values_of_datatype(self.owner.project_id,
                                                               self.datatype_index,
                                                               all_conditions)
        return filtered_datatypes

    def options(self):
        if not self.owner.project_id:
            raise ValueError('A project_id is required in order to query the DB')

        filtered_datatypes = self._get_values_from_db()

        if not self.required:
            choice = None
            yield Option(
                id='{}_{}'.format(self.name, None),
                value=self.missing_value,
                label=str(choice).title(),
                checked=self.data is None
            )

        for i, datatype in enumerate(filtered_datatypes):
            yield Option(
                id='{}_{}'.format(self.name, i),
                value=datatype[2],
                label=self._prepare_display_name(datatype),
                checked=self.data == datatype[2]
            )

        if self.has_all_option:
            if not self.owner.draw_ranges:
                raise ValueError("The owner form should draw ranges inputs in order to support 'All' option")

            all_values = ''
            for fdt in filtered_datatypes:
                all_values += str(fdt[2]) + ','

            choice = "All"
            yield Option(
                id='{}_{}'.format(self.name, choice),
                value=all_values[:-1],
                label=choice,
                checked=self.data is choice
            )

    def get_dt_from_db(self):
        return dao.get_datatype_by_gid(self.data)

    def _prepare_display_name(self, value):
        """
        Populate meta-data fields for data_list (list of DataTypes).

        Private method, to be called recursively.
        It will receive a list of Attributes, and it will populate 'options'
        entry with data references from DB.
        """
        # Here we only populate with DB data, actual
        # XML check will be done after select and submit.
        entity_gid = value[2]
        actual_entity = dao.get_generic_entity(self.datatype_index, entity_gid, "gid")
        display_name = ''
        if actual_entity is not None and len(actual_entity) > 0 and isinstance(actual_entity[0], DataType):
            display_name = actual_entity[0].__class__.__name__
        display_name += ' - ' + (value[3] or "None ")
        if value[5]:
            display_name += ' - From: ' + str(value[5])
        else:
            display_name += utils.date2string(value[4])
        if value[6]:
            display_name += ' - ' + str(value[6])
        display_name += ' - ID:' + str(value[0])

        return display_name

    def _from_post(self):
        if self.unvalidated_data == self.missing_value:
            self.unvalidated_data = None

        if self.required and not self.unvalidated_data:
            raise ValueError('Field required')

        # TODO: ensure is in choices
        try:
            if self.unvalidated_data is None:
                self.data = None
            else:
                self.data = uuid.UUID(self.unvalidated_data)
        except ValueError:
            raise ValueError('The chosen entity does not have a proper GID')



class StrField(TraitField):
    template = 'form_fields/str_field.html'

    def _from_post(self):
        if self.required and (self.unvalidated_data is None or self.unvalidated_data.strip() == ''):
            raise ValueError('Field required')
        self.data = self.unvalidated_data


class BytesField(StrField):
    """ StrField for byte strings. """
    template = 'form_fields/str_field.html'

    def _from_post(self):
        super(BytesField, self)._from_post()
        self.data = self.unvalidated_data.encode('utf-8')


class BoolField(TraitField):
    template = 'form_fields/bool_field.html'

    def _from_post(self):
        self.data = self.unvalidated_data is not None


class IntField(TraitField):
    template = 'form_fields/number_field.html'
    min = None
    max = None

    def _from_post(self):
        super(IntField, self)._from_post()
        self.data = int(self.unvalidated_data)


class FloatField(TraitField):
    template = 'form_fields/number_field.html'
    input_type = "number"
    min = None
    max = None
    step = 'any'

    def _from_post(self):
        super(FloatField, self)._from_post()
        # TODO: Throws exception if attr is optional and has no value
        if self.unvalidated_data and len(self.unvalidated_data) == 0:
            self.unvalidated_data = None
        if self.unvalidated_data:
            self.data = float(self.unvalidated_data)
        else:
            self.data = None


class ArrayField(TraitField):
    template = 'form_fields/str_field.html'

    def _from_post(self):
        data = json.loads(self.unvalidated_data)
        self.data = numpy.array(data, dtype=self.trait_attribute.dtype)

    @property
    def value(self):
        if self.data is None:
            # todo: maybe we need to distinguish None from missing data
            # this None means self.data is missing, either not set or unset cause of validation error
            return self.unvalidated_data
        try:
            if self.data.size > 100:
                data_to_display = self.data[:100]
            else:
                data_to_display = self.data
            return json.dumps(data_to_display.tolist())
        except (TypeError, ValueError):
            return self.unvalidated_data


Option = namedtuple('Option', ['id', 'value', 'label', 'checked'])


class SelectField(TraitField):
    template = 'form_fields/radio_field.html'
    missing_value = 'explicit-None-value'

    def __init__(self, trait_attribute, form, name=None, disabled=False, choices=None, display_none_choice=True):
        super(SelectField, self).__init__(trait_attribute, form, name, disabled)
        if choices:
            self.choices = choices
        else:
            self.choices = {choice: choice for choice in trait_attribute.choices}
        if not self.choices:
            raise ValueError('no choices for field')
        self.display_none_choice = display_none_choice

    @property
    def value(self):
        if self.data is None and not self.trait_attribute.required:
            return self.data
        return super(SelectField, self).value

    def options(self):
        """ to be used from template, assumes self.data is set """
        if self.display_none_choice:
            if not self.trait_attribute.required:
                choice = None
                yield Option(
                    id='{}_{}'.format(self.name, None),
                    value=self.missing_value,
                    label=str(choice).title(),
                    checked=self.data is None
                )

        for i, choice in enumerate(self.choices):
            yield Option(
                id='{}_{}'.format(self.name, i),
                value=choice,
                label=str(choice).title(),
                checked=self.data == self.choices.get(choice)
            )

    # def _from_post(self):
    #     # encode None as a string
    #     if self.unvalidated_data == self.missing_value:
    #         self.unvalidated_data = None
    #
    #     if self.required and not self.unvalidated_data:
    #         raise ValueError('Field required')
    #
    #     if self.unvalidated_data is not None:
    #         # todo muliple values
    #         self.data = self.trait_attribute.field_type(self.unvalidated_data)
    #     else:
    #         self.data = None
    #
    #     allowed = self.trait_attribute.choices
    #     if not self.trait_attribute.required:
    #         allowed = (None,) + allowed
    #
    #     if self.data not in allowed:
    #         raise ValueError('must be one of {}'.format(allowed))

    def fill_from_post(self, post_data):
        super(SelectField, self).fill_from_post(post_data)
        self.data = self.choices.get(self.data)


class MultiSelectField(TraitField):
    template = 'form_fields/checkbox_field.html'

    def __init__(self, trait_attribute, form, name=None, disabled=False):
        super(MultiSelectField, self).__init__(trait_attribute, form, name, disabled)
        if not isinstance(trait_attribute, List):
            raise NotImplementedError('only List in multi select for now')

    def options(self):
        """ to be used from template, assumes self.data is set """
        for i, choice in enumerate(self.trait_attribute.element_choices):
            yield Option(
                id='{}_{}'.format(self.name, i),
                value=choice,
                label=str(choice).title(),
                checked=self.data is not None and choice in self.data
            )

    def _from_post(self):
        if self.unvalidated_data is None:
            if self.required:
                raise ValueError('Field required')
            else:
                return None

        if not isinstance(self.unvalidated_data, list):
            selected = [self.unvalidated_data]
        else:
            selected = self.unvalidated_data

        data = []  # don't mutate self.data until we know all values converted ok

        for s in selected:
            converted_s = self.trait_attribute.element_type(s)
            if converted_s not in self.trait_attribute.element_choices:
                raise ValueError('must be one of {}'.format(self.trait_attribute.element_choices))
            data.append(converted_s)

        self.data = data


# noinspection PyPep8Naming
def ScalarField(trait_attribute, form, name=None, disabled=False):
    # as this makes introspective decisions it has to be moved at a different level
    field_type_for_trait_type = {
        # str: BytesField,
        str: StrField,
        int: IntField,
        float: FloatField,
        bool: BoolField,
    }

    if trait_attribute.choices is not None:
        cls = SelectField
    else:
        cls = field_type_for_trait_type.get(trait_attribute.field_type)

    if isinstance(trait_attribute, List) and trait_attribute.element_choices:
        cls = MultiSelectField

    if cls is None:
        raise ValueError('can not make a scalar field for trait attribute {}'.format(trait_attribute))

    return cls(trait_attribute, form, name=name, disabled=disabled)


class FormField(Field):
    template = 'form_fields/form_field.html'

    def __init__(self, form_class, form, name, label='', doc=''):
        super(FormField, self).__init__(form, name, False, False, label, doc)
        self.form = form_class(prefix=name)

    def fill_from_post(self, post_data):
        self.form.fill_from_post(post_data)
        self.errors = self.form.errors

    def validate(self):
        return self.form.validate()

    def __str__(self):
        return jinja_env.get_template(self.template).render(form=self.form)


class Form(object):
    RANGE_1_NAME = 'range_1'
    RANGE_2_NAME = 'range_2'
    range_1 = None
    range_2 = None

    def __init__(self, prefix='', project_id=None, draw_ranges=True):
        # TODO: makes sense here?
        self.project_id = project_id
        self.prefix = prefix
        self.errors = []
        self.draw_ranges = draw_ranges

    @property
    def fields(self):
        for field in self.__dict__.values():
            if isinstance(field, Field):
                yield field

    @property
    def trait_fields(self):
        for field in self.__dict__.values():
            if isinstance(field, TraitField):
                yield field

    def validate(self):
        valid = True
        for field in self.fields:
            if not field.validate():
                valid = False
        return valid

    def get_errors_dict(self):
        result = {}
        for field in self.fields:
            if not field.validate():
                result[field.name] = field.errors[0]
        return result

    def fill_from_trait(self, trait):
        """
        Sets data for all traited fields from a trait instance.
        Note that FormFields are not TraitFields, so this does not work recursively
        Override to fill in sub-forms
        """
        for field in self.trait_fields:
            f_name = field.trait_attribute.field_name
            if f_name is None:
                # skipp attribute that does not seem to belong to a traited type
                continue
            field.from_trait(trait, f_name)

    def fill_trait(self, datatype):
        """
        Copies the value of the TraitFields to the corresponding Attr-ibutes of the given trait instance
        Note that FormFields are not TraitFields, so this does not work recursively
        Override to fill in sub-forms
        """
        for field in self.trait_fields:
            f_name = field.trait_attribute.field_name
            if f_name is None:
                # skipp attribute that does not seem to belong to a traited type
                continue
            try:
                setattr(datatype, f_name, field.data)
            except TraitError as ex:
                # as field.data is clearly tainted set it to None so that field.unvalidated_data
                # will render and the user can fix the typo's
                field.data = None
                field.errors.append(ex)
                raise

    def fill_from_post(self, form_data):
        for field in self.fields:
            field.fill_from_post(form_data)
        self.range_1 = form_data.get(self.RANGE_1_NAME)
        self.range_2 = form_data.get(self.RANGE_2_NAME)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Change of DB structure from TVB version 1.3.1 to 1.4

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

from sqlalchemy.sql import text
from sqlalchemy import Column, String, Boolean
from migrate.changeset.schema import create_column, drop_column
from tvb.core.entities import model
from tvb.core.entities.storage import SA_SESSIONMAKER
from tvb.basic.logger.builder import get_logger


meta = model.Base.metadata
COL_REG1 = Column('_has_surface_mapping', Boolean, default=False)
COL_REG2 = Column('_has_volume_mapping', Boolean, default=False)
COL_REG3 = Column('_region_mapping', String)
COL_REG4 = Column('_region_mapping_volume', String)
COL_SENSORS = Column('_usable', String)



def upgrade(migrate_engine):
    """
    Upgrade operations go here.
    Don't create your own engine; bind migrate_engine to your metadata.
    """
    meta.bind = migrate_engine
    table1 = meta.tables['MAPPED_TIME_SERIES_DATA']
    table2 = meta.tables['MAPPED_TIME_SERIES_REGION_DATA']
    table3 = meta.tables['MAPPED_SENSORS_DATA']

    create_column(COL_REG1, table1)
    create_column(COL_REG2, table1)
    create_column(COL_REG3, table2)
    create_column(COL_REG4, table2)
    create_column(COL_SENSORS, table3)

    logger = get_logger(__name__)
    sqlite = _exec_update("1", logger)
    if not sqlite:
        _exec_update("True", logger)



def _exec_update(boolean_value, logger):
    session = SA_SESSIONMAKER()
    try:
        logger.info("Executing Db update script 015...")
        session.execute(text("""UPDATE "MAPPED_TIME_SERIES_REGION_DATA" SET _region_mapping =
                        (SELECT dt.gid
                        FROM "MAPPED_REGION_MAPPING_DATA" rm, "DATA_TYPES" dt
                        WHERE dt.id = rm.id AND "MAPPED_TIME_SERIES_REGION_DATA"._connectivity= rm._connectivity);"""))

        session.execute(text("""UPDATE "MAPPED_TIME_SERIES_DATA" SET _has_surface_mapping = """ + boolean_value + """
                            WHERE
                                EXISTS (SELECT * FROM "DATA_TYPES" dt
                                        WHERE dt.id="MAPPED_TIME_SERIES_DATA".id AND dt.type in ('TimeSeriesSurface',
                                                'TimeSeriesEEG', 'TimeSeriesSEEG', 'TimeSeriesMEG'))
                            OR EXISTS (SELECT * from "MAPPED_TIME_SERIES_REGION_DATA" tr
                                    WHERE tr.id="MAPPED_TIME_SERIES_DATA".id AND tr._region_mapping is not NULL);"""))

        session.execute(text("""UPDATE "MAPPED_TIME_SERIES_DATA" SET _has_volume_mapping = """ + boolean_value + """
                            WHERE
                                EXISTS (SELECT * FROM "DATA_TYPES" dt
                                    WHERE dt.id="MAPPED_TIME_SERIES_DATA".id AND dt.type in ('TimeSeriesVolume'));"""))
        session.commit()
        logger.info("DB update script 015 committed.")
        return True

    except Exception as excep:
        logger.exception(excep)
        return False
    finally:
        session.close()



def downgrade(migrate_engine):
    """
    Operations to reverse the above upgrade go here.
    """
    meta.bind = migrate_engine
    table1 = meta.tables['MAPPED_TIME_SERIES_DATA']
    table2 = meta.tables['MAPPED_TIME_SERIES_REGION_DATA']
    table3 = meta.tables['MAPPED_SENSORS_DATA']

    drop_column(COL_REG1, table1)
    drop_column(COL_REG2, table1)
    drop_column(COL_REG3, table2)
    drop_column(COL_REG4, table2)
    drop_column(COL_SENSORS, table3)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

from tvb.core.entities.filters.chain import FilterChain
from tvb.datatypes.sensors import EEG_POLYMORPHIC_IDENTITY as EEG_S
from tvb.datatypes.sensors import MEG_POLYMORPHIC_IDENTITY as MEG_S
from tvb.datatypes.sensors import INTERNAL_POLYMORPHIC_IDENTITY as SEEG_S
from tvb.simulator.monitors import *
from tvb.adapters.simulator.equation_forms import get_ui_name_to_monitor_equation_dict
from tvb.adapters.datatypes.db.region_mapping import RegionMappingIndex
from tvb.adapters.datatypes.db.sensors import SensorsIndex
from tvb.adapters.datatypes.db.surface import SurfaceIndex
from tvb.core.neotraits.forms import Form, ScalarField, ArrayField, DataTypeSelectField, SimpleSelectField


def get_monitor_to_form_dict():
    monitor_class_to_form = {
        Raw: RawMonitorForm,
        SubSample: SubSampleMonitorForm,
        SpatialAverage: SpatialAverageMonitorForm,
        GlobalAverage: GlobalAverageMonitorForm,
        TemporalAverage: TemporalAverageMonitorForm,
        EEG: EEGMonitorForm,
        MEG: MEGMonitorForm,
        iEEG: iEEGMonitorForm,
        Bold: BoldMonitorForm,
        BoldRegionROI: BoldRegionROIMonitorForm
    }

    return monitor_class_to_form


def get_ui_name_to_monitor_dict(surface):
    ui_name_to_monitor = {
        'Raw recording': Raw,
        'Temporally sub-sample': SubSample,
        'Spatial average with temporal sub-sample': SpatialAverage,
        'Global average': GlobalAverage,
        'Temporal average': TemporalAverage,
        'EEG': EEG,
        'MEG': MEG,
        'Intracerebral / Stereo EEG': iEEG,
        'BOLD': Bold
    }

    if surface:
        ui_name_to_monitor['BOLD Region ROI'] = BoldRegionROI

    return ui_name_to_monitor


def get_form_for_monitor(monitor_class):
    return get_monitor_to_form_dict().get(monitor_class)


class MonitorForm(Form):

    def __init__(self, prefix='', project_id=None):
        super(MonitorForm, self).__init__(prefix)
        self.project_id = project_id
        self.period = ScalarField(Monitor.period, self)
        self.variables_of_interest = ArrayField(Monitor.variables_of_interest, self)


class RawMonitorForm(MonitorForm):

    def __init__(self, prefix='', project_id=None):
        super(RawMonitorForm, self).__init__(prefix, project_id)
        self.period = ScalarField(Raw.period, self, disabled=False)
        self.variables_of_interest = ArrayField(Raw.variables_of_interest, self, disabled=False)


class SubSampleMonitorForm(MonitorForm):

    def __init__(self, prefix='', project_id=None):
        super(SubSampleMonitorForm, self).__init__(prefix, project_id)


class SpatialAverageMonitorForm(MonitorForm):

    def __init__(self, prefix='', project_id=None):
        super(SpatialAverageMonitorForm, self).__init__(prefix, project_id)
        self.spatial_mask = ArrayField(SpatialAverage.spatial_mask, self)
        self.default_mask = ScalarField(SpatialAverage.default_mask, self, disabled=True)


class GlobalAverageMonitorForm(MonitorForm):

    def __init__(self, prefix='', project_id=None):
        super(GlobalAverageMonitorForm, self).__init__(prefix, project_id)


class TemporalAverageMonitorForm(Form):

    def __init__(self, prefix='', project_id=None):
        super(TemporalAverageMonitorForm, self).__init__(prefix, project_id)


class ProjectionMonitorForm(MonitorForm):

    def __init__(self, prefix='', project_id=None):
        super(ProjectionMonitorForm, self).__init__(prefix, project_id)
        self.region_mapping = DataTypeSelectField(RegionMappingIndex, self, name='region_mapping', required=True,
                                                  label=Projection.region_mapping.label,
                                                  doc=Projection.region_mapping.doc)
        # self.obsnoise


class EEGMonitorForm(ProjectionMonitorForm):

    def __init__(self, prefix='', project_id=None):
        super(EEGMonitorForm, self).__init__(prefix, project_id)

        sensor_filter = FilterChain(fields=[FilterChain.datatype + '.sensors_type'], operations=["=="],
                                    values=[EEG_S])

        self.projection = DataTypeSelectField(SurfaceIndex, self, name='projection', required=True,
                                              label=EEG.projection.label, doc=EEG.projection.label,
                                              conditions=None)
        self.reference = ScalarField(EEG.reference, self)
        self.sensors = DataTypeSelectField(SensorsIndex, self, name='sensors', required=True, label=EEG.sensors.label,
                                           doc=EEG.sensors.doc, conditions=sensor_filter)
        self.sigma = ScalarField(EEG.sigma, self)


class MEGMonitorForm(ProjectionMonitorForm):

    def __init__(self, prefix='', project_id=None):
        super(MEGMonitorForm, self).__init__(prefix, project_id)

        sensor_filter = FilterChain(fields=[FilterChain.datatype + '.sensors_type'], operations=["=="],
                                    values=[MEG_S])

        self.projection = DataTypeSelectField(SurfaceIndex, self, name='projection', required=True,
                                              label=MEG.projection.label, doc=MEG.projection.doc,
                                              conditions=None)
        self.sensors = DataTypeSelectField(SensorsIndex, self, name='sensors', required=True, label=MEG.sensors.label,
                                           doc=MEG.sensors.doc, conditions=sensor_filter)


class iEEGMonitorForm(ProjectionMonitorForm):

    def __init__(self, prefix='', project_id=None):
        super(iEEGMonitorForm, self).__init__(prefix, project_id)

        sensor_filter = FilterChain(fields=[FilterChain.datatype + '.sensors_type'], operations=["=="],
                                    values=[SEEG_S])

        self.projection = DataTypeSelectField(SurfaceIndex, self, name='projection', required=True,
                                              label=iEEG.projection.label, doc=iEEG.projection.doc,
                                              conditions=None)
        self.sigma = ScalarField(iEEG.sigma, self)
        self.sensors = DataTypeSelectField(SensorsIndex, self, name='sensors', required=True, label=iEEG.sensors.label,
                                           doc=iEEG.sensors.doc, conditions=sensor_filter)


class BoldMonitorForm(MonitorForm):

    def __init__(self, prefix='', project_id=None):
        super(BoldMonitorForm, self).__init__(prefix, project_id)
        self.period = ScalarField(Bold.period, self)
        self.equation_choices = get_ui_name_to_monitor_equation_dict()
        self.equation = SimpleSelectField(self.equation_choices, self, name='equation', required=True, label='Equation')

    def fill_trait(self, datatype):
        super(BoldMonitorForm, self).fill_trait(datatype)
        datatype.period = self.period.data
        datatype.equation = self.equation.data()


class BoldRegionROIMonitorForm(BoldMonitorForm):

    def __init__(self, prefix='', project_id=None):
        super(BoldRegionROIMonitorForm, self).__init__(prefix, project_id)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

from tvb.simulator.models import *
from tvb.adapters.simulator.form_with_ranges import FormWithRanges
from tvb.core.neotraits.forms import Form, ArrayField, MultiSelectField


def get_model_to_form_dict():
    model_class_to_form = {
        Generic2dOscillator: Generic2dOscillatorModelForm,
        Kuramoto: KuramotoModelForm,
        SupHopf: SupHopfModelForm,
        Hopfield: HopfieldModelForm,
        Epileptor: EpileptorModelForm,
        Epileptor2D: Epileptor2DModelForm,
        EpileptorCodim3: EpileptorCodim3ModelForm,
        EpileptorCodim3SlowMod: EpileptorCodim3SlowModModelForm,
        EpileptorRestingState: EpileptorRestingStateModelForm,
        JansenRit: JansenRitModelForm,
        ZetterbergJansen: ZetterbergJansenModelForm,
        ReducedWongWang: ReducedWongWangModelForm,
        ReducedWongWangExcInh: ReducedWongWangExcInhModelForm,
        ReducedSetFitzHughNagumo: ReducedSetFitzHughNagumoModelForm,
        ReducedSetHindmarshRose: ReducedSetHindmarshRoseModelForm,
        ZerlautFirstOrder: ZerlautFirstOrderModelForm,
        ZerlautSecondOrder: ZerlautSecondOrderModelForm,
        Linear: LinearModelForm,
        WilsonCowan: WilsonCowanModelForm,
        LarterBreakspear: LarterBreakspearModelForm
    }

    return model_class_to_form


def get_ui_name_to_model():
    ui_name_to_model = {
        'Generic 2d Oscillator': Generic2dOscillator,
        'Kuramoto Oscillator': Kuramoto,
        'supHopf': SupHopf,
        'Hopfield': Hopfield,
        'Epileptor': Epileptor,
        'Epileptor2D': Epileptor2D,
        'Epileptor codim 3': EpileptorCodim3,
        'Epileptor codim 3 ultra-slow modulations': EpileptorCodim3SlowMod,
        'EpileptorRestingState': EpileptorRestingState,
        'Jansen-Rit': JansenRit,
        'Zetterberg-Jansen': ZetterbergJansen,
        'Reduced Wong-Wang': ReducedWongWang,
        'Reduced Wong-Wang with Excitatory and Inhibitory Coupled Populations': ReducedWongWangExcInh,
        'Stefanescu-Jirsa 2D': ReducedSetFitzHughNagumo,
        'Stefanescu-Jirsa 3D': ReducedSetHindmarshRose,
        'Zerlaut adaptation first order': ZerlautFirstOrder,
        'Zerlaut adaptation second order': ZerlautSecondOrder,
        'Linear model': Linear,
        'Wilson-Cowan': WilsonCowan,
        'Larter-Breakspear': LarterBreakspear
    }
    return ui_name_to_model


def get_form_for_model(model_class):
    return get_model_to_form_dict().get(model_class)


class StateVariableRangesForm(Form):

    def __init__(self, prefix=''):
        super(StateVariableRangesForm, self).__init__(prefix)


class Generic2dOscillatorModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(Generic2dOscillatorModelForm, self).__init__(prefix)
        self.tau = ArrayField(Generic2dOscillator.tau, self)
        self.I = ArrayField(Generic2dOscillator.I, self)
        self.a = ArrayField(Generic2dOscillator.a, self)
        self.b = ArrayField(Generic2dOscillator.b, self)
        self.c = ArrayField(Generic2dOscillator.c, self)
        self.d = ArrayField(Generic2dOscillator.d, self)
        self.e = ArrayField(Generic2dOscillator.e, self)
        self.f = ArrayField(Generic2dOscillator.f, self)
        self.g = ArrayField(Generic2dOscillator.g, self)
        self.alpha = ArrayField(Generic2dOscillator.alpha, self)
        self.beta = ArrayField(Generic2dOscillator.beta, self)
        self.gamma = ArrayField(Generic2dOscillator.gamma, self)
        self.variables_of_interest = MultiSelectField(Generic2dOscillator.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['tau', 'a', 'b', 'c', 'I', 'd', 'e', 'f', 'g', 'alpha', 'beta', 'gamma']


class KuramotoModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(KuramotoModelForm, self).__init__(prefix)
        self.omega = ArrayField(Kuramoto.omega, self)
        self.variables_of_interest = MultiSelectField(Kuramoto.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['omega']


class SupHopfModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(SupHopfModelForm, self).__init__(prefix)
        self.a = ArrayField(SupHopf.a, self)
        self.omega = ArrayField(SupHopf.omega, self)
        self.variables_of_interest = MultiSelectField(SupHopf.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['a', 'omega']


class HopfieldModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(HopfieldModelForm, self).__init__(prefix)
        self.taux = ArrayField(Hopfield.taux, self)
        self.tauT = ArrayField(Hopfield.tauT, self)
        self.dynamic = ArrayField(Hopfield.dynamic, self)
        self.variables_of_interest = MultiSelectField(Hopfield.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['taux', 'tauT', 'dynamic']


class EpileptorModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(EpileptorModelForm, self).__init__(prefix)
        self.a = ArrayField(Epileptor.a, self)
        self.b = ArrayField(Epileptor.b, self)
        self.c = ArrayField(Epileptor.c, self)
        self.d = ArrayField(Epileptor.d, self)
        self.r = ArrayField(Epileptor.r, self)
        self.s = ArrayField(Epileptor.s, self)
        self.x0 = ArrayField(Epileptor.x0, self)
        self.Iext = ArrayField(Epileptor.Iext, self)
        self.slope = ArrayField(Epileptor.slope, self)
        self.Iext2 = ArrayField(Epileptor.Iext2, self)
        self.tau = ArrayField(Epileptor.tau, self)
        self.aa = ArrayField(Epileptor.aa, self)
        self.bb = ArrayField(Epileptor.bb, self)
        self.Kvf = ArrayField(Epileptor.Kvf, self)
        self.Kf = ArrayField(Epileptor.Kf, self)
        self.Ks = ArrayField(Epileptor.Ks, self)
        self.tt = ArrayField(Epileptor.tt, self)
        self.modification = ArrayField(Epileptor.modification, self)
        self.variables_of_interest = MultiSelectField(Epileptor.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ["Iext", "Iext2", "r", "x0", "slope"]


class Epileptor2DModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(Epileptor2DModelForm, self).__init__(prefix)
        self.a = ArrayField(Epileptor2D.a, self)
        self.b = ArrayField(Epileptor2D.b, self)
        self.c = ArrayField(Epileptor2D.c, self)
        self.d = ArrayField(Epileptor2D.d, self)
        self.r = ArrayField(Epileptor2D.r, self)
        self.x0 = ArrayField(Epileptor2D.x0, self)
        self.Iext = ArrayField(Epileptor2D.Iext, self)
        self.slope = ArrayField(Epileptor2D.slope, self)
        self.Kvf = ArrayField(Epileptor2D.Kvf, self)
        self.Ks = ArrayField(Epileptor2D.Ks, self)
        self.tt = ArrayField(Epileptor2D.tt, self)
        self.modification = ArrayField(Epileptor2D.modification, self)
        self.variables_of_interest = MultiSelectField(Epileptor2D.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ["r", "Iext", "x0"]


class EpileptorCodim3ModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(EpileptorCodim3ModelForm, self).__init__(prefix)
        self.mu1_start = ArrayField(EpileptorCodim3.mu1_start, self)
        self.mu2_start = ArrayField(EpileptorCodim3.mu2_start, self)
        self.nu_start = ArrayField(EpileptorCodim3.nu_start, self)
        self.mu1_stop = ArrayField(EpileptorCodim3.mu1_stop, self)
        self.mu2_stop = ArrayField(EpileptorCodim3.mu2_stop, self)
        self.nu_stop = ArrayField(EpileptorCodim3.nu_stop, self)
        self.b = ArrayField(EpileptorCodim3.b, self)
        self.R = ArrayField(EpileptorCodim3.R, self)
        self.c = ArrayField(EpileptorCodim3.c, self)
        self.dstar = ArrayField(EpileptorCodim3.dstar, self)
        self.Ks = ArrayField(EpileptorCodim3.Ks, self)
        self.N = ArrayField(EpileptorCodim3.N, self)
        self.modification = ArrayField(EpileptorCodim3.modification, self)
        self.variables_of_interest = MultiSelectField(EpileptorCodim3.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['mu1_start', 'mu2_start', 'nu_start', 'mu1_stop', 'mu2_stop', 'nu_stop', 'b', 'R', 'c', 'dstar', 'N',
                'Ks']


class EpileptorCodim3SlowModModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(EpileptorCodim3SlowModModelForm, self).__init__(prefix)
        self.mu1_Ain = ArrayField(EpileptorCodim3SlowMod.mu1_Ain, self)
        self.mu2_Ain = ArrayField(EpileptorCodim3SlowMod.mu2_Ain, self)
        self.nu_Ain = ArrayField(EpileptorCodim3SlowMod.nu_Ain, self)
        self.mu1_Bin = ArrayField(EpileptorCodim3SlowMod.mu1_Bin, self)
        self.mu2_Bin = ArrayField(EpileptorCodim3SlowMod.mu2_Bin, self)
        self.nu_Bin = ArrayField(EpileptorCodim3SlowMod.nu_Bin, self)
        self.mu1_Aend = ArrayField(EpileptorCodim3SlowMod.mu1_Aend, self)
        self.mu2_Aend = ArrayField(EpileptorCodim3SlowMod.mu2_Aend, self)
        self.nu_Aend = ArrayField(EpileptorCodim3SlowMod.nu_Aend, self)
        self.mu1_Bend = ArrayField(EpileptorCodim3SlowMod.mu1_Bend, self)
        self.mu2_Bend = ArrayField(EpileptorCodim3SlowMod.mu2_Bend, self)
        self.nu_Bend = ArrayField(EpileptorCodim3SlowMod.nu_Bend, self)
        self.b = ArrayField(EpileptorCodim3SlowMod.b, self)
        self.R = ArrayField(EpileptorCodim3SlowMod.R, self)
        self.c = ArrayField(EpileptorCodim3SlowMod.c, self)
        self.cA = ArrayField(EpileptorCodim3SlowMod.cA, self)
        self.cB = ArrayField(EpileptorCodim3SlowMod.cB, self)
        self.dstar = ArrayField(EpileptorCodim3SlowMod.dstar, self)
        self.Ks = ArrayField(EpileptorCodim3SlowMod.Ks, self)
        self.N = ArrayField(EpileptorCodim3SlowMod.N, self)
        self.modification = ArrayField(EpileptorCodim3SlowMod.modification, self)
        self.variables_of_interest = ArrayField(EpileptorCodim3SlowMod.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['mu1_Ain', 'mu2_Ain', 'nu_Ain', 'mu1_Bin', 'mu2_Bin', 'nu_Bin', 'mu1_Aend', 'mu2_Aend', 'nu_Aend',
                'mu1_Bend', 'mu2_Bend', 'nu_Bend', 'b', 'R', 'c', 'dstar', 'N']


class EpileptorRestingStateModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(EpileptorRestingStateModelForm, self).__init__(prefix)
        self.a = ArrayField(EpileptorRestingState.a, self)
        self.b = ArrayField(EpileptorRestingState.b, self)
        self.c = ArrayField(EpileptorRestingState.c, self)
        self.d = ArrayField(EpileptorRestingState.d, self)
        self.r = ArrayField(EpileptorRestingState.r, self)
        self.s = ArrayField(EpileptorRestingState.s, self)
        self.x0 = ArrayField(EpileptorRestingState.x0, self)
        self.Iext = ArrayField(EpileptorRestingState.Iext, self)
        self.slope = ArrayField(EpileptorRestingState.slope, self)
        self.Iext2 = ArrayField(EpileptorRestingState.Iext2, self)
        self.tau = ArrayField(EpileptorRestingState.tau, self)
        self.aa = ArrayField(EpileptorRestingState.aa, self)
        self.bb = ArrayField(EpileptorRestingState.bb, self)
        self.Kvf = ArrayField(EpileptorRestingState.Kvf, self)
        self.Kf = ArrayField(EpileptorRestingState.Kf, self)
        self.Ks = ArrayField(EpileptorRestingState.Ks, self)
        self.tt = ArrayField(EpileptorRestingState.tt, self)
        self.tau_rs = ArrayField(EpileptorRestingState.tau_rs, self)
        self.I_rs = ArrayField(EpileptorRestingState.I_rs, self)
        self.a_rs = ArrayField(EpileptorRestingState.a_rs, self)
        self.b_rs = ArrayField(EpileptorRestingState.b_rs, self)
        self.d_rs = ArrayField(EpileptorRestingState.d_rs, self)
        self.e_rs = ArrayField(EpileptorRestingState.e_rs, self)
        self.f_rs = ArrayField(EpileptorRestingState.f_rs, self)
        self.alpha_rs = ArrayField(EpileptorRestingState.alpha_rs, self)
        self.beta_rs = ArrayField(EpileptorRestingState.beta_rs, self)
        self.K_rs = ArrayField(EpileptorRestingState.K_rs, self)
        self.p = ArrayField(EpileptorRestingState.p, self)
        self.variables_of_interest = MultiSelectField(EpileptorRestingState.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['Iext', 'Iext2', 'r', 'x0', 'slope', 'tau_rs', 'a_rs', 'b_rs', 'I_rs', 'd_rs', 'e_rs', 'f_rs',
                'alpha_rs', 'beta_rs', 'gamma_rs']


class JansenRitModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(JansenRitModelForm, self).__init__(prefix)
        self.A = ArrayField(JansenRit.A, self)
        self.B = ArrayField(JansenRit.B, self)
        self.a = ArrayField(JansenRit.a, self)
        self.b = ArrayField(JansenRit.b, self)
        self.v0 = ArrayField(JansenRit.v0, self)
        self.nu_max = ArrayField(JansenRit.nu_max, self)
        self.r = ArrayField(JansenRit.r, self)
        self.J = ArrayField(JansenRit.J, self)
        self.a_1 = ArrayField(JansenRit.a_1, self)
        self.a_2 = ArrayField(JansenRit.a_2, self)
        self.a_3 = ArrayField(JansenRit.a_3, self)
        self.a_4 = ArrayField(JansenRit.a_4, self)
        self.p_min = ArrayField(JansenRit.p_min, self)
        self.p_max = ArrayField(JansenRit.p_max, self)
        self.mu = ArrayField(JansenRit.mu, self)
        self.variables_of_interest = ArrayField(JansenRit.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['A', 'B', 'a', 'b', 'v0', 'nu_max', 'r', 'J', 'a_1', 'a_2', 'a_3', 'a_4', 'p_min', 'p_max', 'mu']


class ZetterbergJansenModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(ZetterbergJansenModelForm, self).__init__(prefix)
        self.He = ArrayField(ZetterbergJansen.He, self)
        self.Hi = ArrayField(ZetterbergJansen.Hi, self)
        self.ke = ArrayField(ZetterbergJansen.ke, self)
        self.ki = ArrayField(ZetterbergJansen.ki, self)
        self.e0 = ArrayField(ZetterbergJansen.e0, self)
        self.rho_2 = ArrayField(ZetterbergJansen.rho_2, self)
        self.rho_1 = ArrayField(ZetterbergJansen.rho_1, self)
        self.gamma_1 = ArrayField(ZetterbergJansen.gamma_1, self)
        self.gamma_2 = ArrayField(ZetterbergJansen.gamma_2, self)
        self.gamma_3 = ArrayField(ZetterbergJansen.gamma_3, self)
        self.gamma_4 = ArrayField(ZetterbergJansen.gamma_4, self)
        self.gamma_5 = ArrayField(ZetterbergJansen.gamma_5, self)
        self.gamma_1T = ArrayField(ZetterbergJansen.gamma_1T, self)
        self.gamma_2T = ArrayField(ZetterbergJansen.gamma_2T, self)
        self.gamma_3T = ArrayField(ZetterbergJansen.gamma_3T, self)
        self.P = ArrayField(ZetterbergJansen.P, self)
        self.U = ArrayField(ZetterbergJansen.U, self)
        self.Q = ArrayField(ZetterbergJansen.Q, self)
        self.variables_of_interest = MultiSelectField(ZetterbergJansen.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['He', 'Hi', 'ke', 'ki', 'e0', 'rho_2', 'rho_1', 'gamma_1', 'gamma_2', 'gamma_3', 'gamma_4', 'gamma_5',
                'P', 'U', 'Q']


class ReducedWongWangModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(ReducedWongWangModelForm, self).__init__(prefix)
        self.a = ArrayField(ReducedWongWang.a, self)
        self.b = ArrayField(ReducedWongWang.b, self)
        self.d = ArrayField(ReducedWongWang.d, self)
        self.gamma = ArrayField(ReducedWongWang.gamma, self)
        self.tau_s = ArrayField(ReducedWongWang.tau_s, self)
        self.w = ArrayField(ReducedWongWang.w, self)
        self.J_N = ArrayField(ReducedWongWang.J_N, self)
        self.I_o = ArrayField(ReducedWongWang.I_o, self)
        self.sigma_noise = ArrayField(ReducedWongWang.sigma_noise, self)
        self.variables_of_interest = MultiSelectField(ReducedWongWang.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['a', 'b', 'd', 'gamma', 'tau_s', 'w', 'J_N', 'I_o']


class ReducedWongWangExcInhModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(ReducedWongWangExcInhModelForm, self).__init__(prefix)
        self.a_e = ArrayField(ReducedWongWangExcInh.a_e, self)
        self.b_e = ArrayField(ReducedWongWangExcInh.b_e, self)
        self.d_e = ArrayField(ReducedWongWangExcInh.d_e, self)
        self.gamma_e = ArrayField(ReducedWongWangExcInh.gamma_e, self)
        self.tau_e = ArrayField(ReducedWongWangExcInh.tau_e, self)
        self.w_p = ArrayField(ReducedWongWangExcInh.w_p, self)
        self.J_N = ArrayField(ReducedWongWangExcInh.J_N, self)
        self.W_e = ArrayField(ReducedWongWangExcInh.W_e, self)
        self.a_i = ArrayField(ReducedWongWangExcInh.a_i, self)
        self.b_i = ArrayField(ReducedWongWangExcInh.b_i, self)
        self.d_i = ArrayField(ReducedWongWangExcInh.d_i, self)
        self.gamma_i = ArrayField(ReducedWongWangExcInh.gamma_i, self)
        self.tau_i = ArrayField(ReducedWongWangExcInh.tau_i, self)
        self.J_i = ArrayField(ReducedWongWangExcInh.J_i, self)
        self.W_i = ArrayField(ReducedWongWangExcInh.W_i, self)
        self.I_o = ArrayField(ReducedWongWangExcInh.I_o, self)
        self.G = ArrayField(ReducedWongWangExcInh.G, self)
        self.lamda = ArrayField(ReducedWongWangExcInh.lamda, self)
        self.variables_of_interest = MultiSelectField(ReducedWongWangExcInh.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['a_e', 'b_e', 'd_e', 'gamma_e', 'tau_e', 'W_e', 'w_p', 'J_N', 'a_i', 'b_i', 'd_i', 'gamma_i', 'tau_i',
                'W_i', 'J_i', 'I_o', 'G', 'lamda']


class ReducedSetFitzHughNagumoModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(ReducedSetFitzHughNagumoModelForm, self).__init__(prefix)
        self.tau = ArrayField(ReducedSetFitzHughNagumo.tau, self)
        self.a = ArrayField(ReducedSetFitzHughNagumo.a, self)
        self.b = ArrayField(ReducedSetFitzHughNagumo.b, self)
        self.K11 = ArrayField(ReducedSetFitzHughNagumo.K11, self)
        self.K12 = ArrayField(ReducedSetFitzHughNagumo.K12, self)
        self.K21 = ArrayField(ReducedSetFitzHughNagumo.K21, self)
        self.sigma = ArrayField(ReducedSetFitzHughNagumo.sigma, self)
        self.mu = ArrayField(ReducedSetFitzHughNagumo.mu, self)
        self.variables_of_interest = MultiSelectField(ReducedSetFitzHughNagumo.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['tau', 'a', 'b', 'K11', 'K12', 'K21', 'sigma', 'mu']


class ReducedSetHindmarshRoseModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(ReducedSetHindmarshRoseModelForm, self).__init__(prefix)
        self.r = ArrayField(ReducedSetHindmarshRose.r, self)
        self.a = ArrayField(ReducedSetHindmarshRose.a, self)
        self.b = ArrayField(ReducedSetHindmarshRose.b, self)
        self.c = ArrayField(ReducedSetHindmarshRose.c, self)
        self.d = ArrayField(ReducedSetHindmarshRose.d, self)
        self.s = ArrayField(ReducedSetHindmarshRose.s, self)
        self.xo = ArrayField(ReducedSetHindmarshRose.xo, self)
        self.K11 = ArrayField(ReducedSetHindmarshRose.K11, self)
        self.K12 = ArrayField(ReducedSetHindmarshRose.K12, self)
        self.K21 = ArrayField(ReducedSetHindmarshRose.K21, self)
        self.sigma = ArrayField(ReducedSetHindmarshRose.sigma, self)
        self.mu = ArrayField(ReducedSetHindmarshRose.mu, self)
        self.variables_of_interest = MultiSelectField(ReducedSetHindmarshRose.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['r', 'a', 'b', 'c', 'd', 's', 'xo', 'K11', 'K12', 'K21', 'sigma', 'mu']


class ZerlautFirstOrderModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(ZerlautFirstOrderModelForm, self).__init__(prefix)
        self.g_L = ArrayField(ZerlautFirstOrder.g_L, self)
        self.E_L_e = ArrayField(ZerlautFirstOrder.E_L_e, self)
        self.E_L_i = ArrayField(ZerlautFirstOrder.E_L_i, self)
        self.C_m = ArrayField(ZerlautFirstOrder.C_m, self)
        self.b = ArrayField(ZerlautFirstOrder.b, self)
        self.tau_w = ArrayField(ZerlautFirstOrder.tau_w, self)
        self.E_e = ArrayField(ZerlautFirstOrder.E_e, self)
        self.E_i = ArrayField(ZerlautFirstOrder.E_i, self)
        self.Q_e = ArrayField(ZerlautFirstOrder.Q_e, self)
        self.Q_i = ArrayField(ZerlautFirstOrder.Q_i, self)
        self.tau_e = ArrayField(ZerlautFirstOrder.tau_e, self)
        self.tau_i = ArrayField(ZerlautFirstOrder.tau_i, self)
        self.N_tot = ArrayField(ZerlautFirstOrder.N_tot, self)
        self.p_connect = ArrayField(ZerlautFirstOrder.p_connect, self)
        self.g = ArrayField(ZerlautFirstOrder.g, self)
        self.T = ArrayField(ZerlautFirstOrder.T, self)
        self.P_e = ArrayField(ZerlautFirstOrder.P_e, self)
        self.P_i = ArrayField(ZerlautFirstOrder.P_i, self)
        self.external_input = ArrayField(ZerlautFirstOrder.external_input, self)
        self.variables_of_interest = MultiSelectField(ZerlautFirstOrder.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['g_L', 'E_L_e', 'E_L_i', 'C_m', 'b', 'tau_w', 'E_e', 'E_i', 'Q_e', 'Q_i', 'tau_e', 'tau_i', 'N_tot',
                'p_connect', 'g', 'T', 'external_input']


class ZerlautSecondOrderModelForm(ZerlautFirstOrderModelForm):

    def __init__(self, prefix=''):
        super(ZerlautSecondOrderModelForm, self).__init__(prefix)
        self.variables_of_interest = MultiSelectField(ZerlautSecondOrder.variables_of_interest, self)


class LinearModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(LinearModelForm, self).__init__(prefix)
        self.gamma = ArrayField(Linear.gamma, self)
        self.variables_of_interest = MultiSelectField(Linear.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['gamma']


class WilsonCowanModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(WilsonCowanModelForm, self).__init__(prefix)
        self.c_ee = ArrayField(WilsonCowan.c_ee, self)
        self.c_ie = ArrayField(WilsonCowan.c_ie, self)
        self.c_ei = ArrayField(WilsonCowan.c_ei, self)
        self.c_ii = ArrayField(WilsonCowan.c_ii, self)
        self.tau_e = ArrayField(WilsonCowan.tau_e, self)
        self.tau_i = ArrayField(WilsonCowan.tau_i, self)
        self.a_e = ArrayField(WilsonCowan.a_e, self)
        self.b_e = ArrayField(WilsonCowan.b_e, self)
        self.c_e = ArrayField(WilsonCowan.c_e, self)
        self.theta_e = ArrayField(WilsonCowan.theta_e, self)
        self.a_i = ArrayField(WilsonCowan.a_i, self)
        self.b_i = ArrayField(WilsonCowan.b_i, self)
        self.theta_i = ArrayField(WilsonCowan.theta_i, self)
        self.c_i = ArrayField(WilsonCowan.c_i, self)
        self.r_e = ArrayField(WilsonCowan.r_e, self)
        self.r_i = ArrayField(WilsonCowan.r_i, self)
        self.k_e = ArrayField(WilsonCowan.k_e, self)
        self.k_i = ArrayField(WilsonCowan.k_i, self)
        self.P = ArrayField(WilsonCowan.P, self)
        self.Q = ArrayField(WilsonCowan.Q, self)
        self.alpha_e = ArrayField(WilsonCowan.alpha_e, self)
        self.alpha_i = ArrayField(WilsonCowan.alpha_i, self)
        self.variables_of_interest = MultiSelectField(WilsonCowan.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['c_ee', 'c_ei', 'c_ie', 'c_ii', 'tau_e', 'tau_i', 'a_e', 'b_e', 'c_e', 'a_i', 'b_i', 'c_i', 'r_e',
                'r_i', 'k_e', 'k_i', 'P', 'Q', 'theta_e', 'theta_i', 'alpha_e', 'alpha_i']


class LarterBreakspearModelForm(FormWithRanges):

    def __init__(self, prefix=''):
        super(LarterBreakspearModelForm, self).__init__(prefix)
        self.gCa = ArrayField(LarterBreakspear.gCa, self)
        self.gK = ArrayField(LarterBreakspear.gK, self)
        self.gL = ArrayField(LarterBreakspear.gL, self)
        self.phi = ArrayField(LarterBreakspear.phi, self)
        self.gNa = ArrayField(LarterBreakspear.gNa, self)
        self.TK = ArrayField(LarterBreakspear.TK, self)
        self.TCa = ArrayField(LarterBreakspear.TCa, self)
        self.TNa = ArrayField(LarterBreakspear.TNa, self)
        self.VCa = ArrayField(LarterBreakspear.VCa, self)
        self.VK = ArrayField(LarterBreakspear.VK, self)
        self.VL = ArrayField(LarterBreakspear.VL, self)
        self.VNa = ArrayField(LarterBreakspear.VNa, self)
        self.d_K = ArrayField(LarterBreakspear.d_K, self)
        self.tau_K = ArrayField(LarterBreakspear.tau_K, self)
        self.d_Na = ArrayField(LarterBreakspear.d_Na, self)
        self.d_Ca = ArrayField(LarterBreakspear.d_Ca, self)
        self.aei = ArrayField(LarterBreakspear.aei, self)
        self.aie = ArrayField(LarterBreakspear.aie, self)
        self.b = ArrayField(LarterBreakspear.b, self)
        self.C = ArrayField(LarterBreakspear.C, self)
        self.ane = ArrayField(LarterBreakspear.ane, self)
        self.ani = ArrayField(LarterBreakspear.ani, self)
        self.aee = ArrayField(LarterBreakspear.aee, self)
        self.Iext = ArrayField(LarterBreakspear.Iext, self)
        self.rNMDA = ArrayField(LarterBreakspear.rNMDA, self)
        self.VT = ArrayField(LarterBreakspear.VT, self)
        self.d_V = ArrayField(LarterBreakspear.d_V, self)
        self.ZT = ArrayField(LarterBreakspear.ZT, self)
        self.d_Z = ArrayField(LarterBreakspear.d_Z, self)
        self.QV_max = ArrayField(LarterBreakspear.QV_max, self)
        self.QZ_max = ArrayField(LarterBreakspear.QZ_max, self)
        self.t_scale = ArrayField(LarterBreakspear.t_scale, self)
        self.variables_of_interest = MultiSelectField(LarterBreakspear.variables_of_interest, self)

    @staticmethod
    def get_params_configurable_in_phase_plane():
        return ['gCa', 'gK', 'gL', 'phi', 'gNa', 'TK', 'TCa', 'TNa', 'VCa', 'VK', 'VL', 'VNa', 'd_K', 'tau_K', 'd_Na',
                'd_Ca', 'aei', 'aie', 'b', 'C', 'ane', 'ani', 'aee', 'Iext', 'rNMDA', 'VT', 'd_V', 'ZT', 'd_Z',
                'QV_max', 'QZ_max', 't_scale']



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
import json
from sqlalchemy import Column, Integer, ForeignKey, String, Float
from sqlalchemy.orm import relationship
from tvb.datatypes.temporal_correlations import CrossCorrelation
from tvb.adapters.datatypes.db.time_series import TimeSeriesIndex
from tvb.core.entities.model.model_datatype import DataType
from tvb.core.neotraits.db import from_ndarray


class CrossCorrelationIndex(DataType):
    id = Column(Integer, ForeignKey(DataType.id), primary_key=True)

    array_data_min = Column(Float)
    array_data_max = Column(Float)
    array_data_mean = Column(Float)

    source_gid = Column(String(32), ForeignKey(TimeSeriesIndex.gid), nullable=not CrossCorrelation.source.required)
    source = relationship(TimeSeriesIndex, foreign_keys=source_gid, primaryjoin=TimeSeriesIndex.gid == source_gid)

    labels_ordering = Column(String, nullable=False)
    subtype = Column(String)

    def fill_from_has_traits(self, datatype):
        # type: (CrossCorrelation)  -> None
        super(CrossCorrelationIndex, self).fill_from_has_traits(datatype)
        self.array_data_min, self.array_data_max, self.array_data_mean = from_ndarray(datatype.array_data)
        self.labels_ordering = json.dumps(datatype.labels_ordering)
        self.subtype = datatype.__class__.__name__
        self.source_gid = datatype.source.gid.hex
